"""Geometry utilities for KotOR model processing.

This module provides reusable geometry operations that are not specific to any
particular rendering backend. These utilities can be used by any rendering
backend implementation.

References:
----------
        Based on swkotor.exe GFF structure:
        - CResGFF::CreateGFFFile @ 0x00411260 - Creates GFF file structure
        - Vector3/Vector4 structures used in GFF fields
        - Libraries/PyKotor/src/pykotor/resource/formats/mdl/io_mdl.py:1448-1577 - Tangent space calculation
    Derivations and Other Implementations:
    ----------
        - https://github.com/th3w1zard1/mdlops/tree/master/MDLOpsM.pm:5470-5596


"""

from __future__ import annotations

import math

from typing import TYPE_CHECKING, Any, overload

from pykotor.resource.formats.mdl import MDLSkin

# Runtime imports - these are used in the function bodies
from utility.common.geometry import Matrix4, Vector3, Vector4

if TYPE_CHECKING:
    from numpy import ndarray

    from pykotor.resource.formats.mdl import MDLMesh


def compute_per_vertex_tangent_space(
    mesh: MDLMesh,
) -> dict[int, tuple[Vector3, Vector3]]:
    """Compute per-vertex tangent and binormal vectors for a mesh.

    This function computes tangent space basis vectors for each vertex by:
    1. Computing tangent/bitangent for each face
    2. Accumulating contributions from all faces sharing each vertex
    3. Averaging and normalizing the results

    Args:
    ----
        mesh: MDL mesh with vertex positions, UVs, and faces

    Returns:
    -------
        Dictionary mapping vertex index to (tangent, binormal) tuple

    References:
    ----------
        Based on swkotor.exe GFF structure:
        - CResGFF::CreateGFFFile @ 0x00411260 - Creates GFF file structure
        - Vector3/Vector4 structures used in GFF fields
        - Libraries/PyKotor/src/pykotor/resource/formats/mdl/io_mdl.py:1448-1577 - Tangent space calculation
        - https://github.com/th3w1zard1/mdlops/tree/master/MDLOpsM.pm:5470-5596
    """
    from pykotor.resource.formats.mdl.io_mdl import _calculate_face_normal, _calculate_tangent_space

    vertex_tangents: dict[int, list[Vector3]] = {}
    vertex_binormals: dict[int, list[Vector3]] = {}

    # Compute per-face tangent space
    for face in mesh.faces:
        v1 = mesh.vertex_positions[face.v1]
        v2 = mesh.vertex_positions[face.v2]
        v3 = mesh.vertex_positions[face.v3]

        if face.v1 >= len(mesh.vertex_uvs) or face.v2 >= len(mesh.vertex_uvs) or face.v3 >= len(mesh.vertex_uvs):
            continue

        uv1 = mesh.vertex_uvs[face.v1]
        uv2 = mesh.vertex_uvs[face.v2]
        uv3 = mesh.vertex_uvs[face.v3]

        face_normal, _ = _calculate_face_normal(v1, v2, v3)
        tangent, binormal = _calculate_tangent_space(v1, v2, v3, uv1, uv2, uv3, face_normal)

        # Accumulate for each vertex
        for v_idx in [face.v1, face.v2, face.v3]:
            if v_idx not in vertex_tangents:
                vertex_tangents[v_idx] = []
                vertex_binormals[v_idx] = []
            vertex_tangents[v_idx].append(tangent)
            vertex_binormals[v_idx].append(binormal)

    # Average accumulated tangents/binormals
    result: dict[int, tuple[Vector3, Vector3]] = {}
    for v_idx in vertex_tangents:
        # Average tangents
        avg_tangent = Vector3(0, 0, 0)
        for t in vertex_tangents[v_idx]:
            avg_tangent = Vector3(avg_tangent.x + t.x, avg_tangent.y + t.y, avg_tangent.z + t.z)
        count = len(vertex_tangents[v_idx])
        avg_tangent = Vector3(avg_tangent.x / count, avg_tangent.y / count, avg_tangent.z / count)

        # Normalize
        length = (avg_tangent.x**2 + avg_tangent.y**2 + avg_tangent.z**2) ** 0.5
        if length > 0:
            avg_tangent = Vector3(avg_tangent.x / length, avg_tangent.y / length, avg_tangent.z / length)

        # Average binormals
        avg_binormal = Vector3(0, 0, 0)
        for b in vertex_binormals[v_idx]:
            avg_binormal = Vector3(avg_binormal.x + b.x, avg_binormal.y + b.y, avg_binormal.z + b.z)
        avg_binormal = Vector3(avg_binormal.x / count, avg_binormal.y / count, avg_binormal.z / count)

        # Normalize
        length = (avg_binormal.x**2 + avg_binormal.y**2 + avg_binormal.z**2) ** 0.5
        if length > 0:
            avg_binormal = Vector3(avg_binormal.x / length, avg_binormal.y / length, avg_binormal.z / length)

        result[v_idx] = (avg_tangent, avg_binormal)

    return result


def determine_vertex_format_requirements(mesh: MDLMesh) -> dict[str, bool]:
    """Determine what vertex format attributes are needed for a mesh.

    This is a backend-agnostic way to determine vertex format requirements
    that can be used by any rendering backend implementation.

    Args:
    ----
        mesh: MDL mesh to analyze

    Returns:
    -------
        Dictionary with boolean flags for each attribute type:
        - has_normals: True if mesh has vertex normals
        - has_tangent_space: True if tangent space should be computed
        - has_lightmap: True if mesh has lightmap UVs
        - has_skinning: True if mesh has bone weights
        - has_uv2: True if mesh has second UV set

    References:
    ----------
        Based on swkotor.exe GFF structure:
        - CResGFF::CreateGFFFile @ 0x00411260 - Creates GFF file structure
        - Vector3/Vector4 structures used in GFF fields

    Derivations and Other Implementations:
    ----------
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/three/odyssey/OdysseyModel3D.ts:1169-1197


    """
    return {
        "has_normals": len(mesh.vertex_normals) > 0 if mesh.vertex_normals is not None else False,
        "has_tangent_space": (len(mesh.vertex_normals) > 0 if mesh.vertex_normals is not None else False) and len(mesh.faces) > 0 and len(mesh.vertex_uvs) > 0,
        "has_lightmap": mesh.has_lightmap and (len(mesh.vertex_uv2) > 0 if mesh.vertex_uv2 is not None else False),
        "has_skinning": isinstance(mesh, MDLSkin) and mesh is not None,
        "has_uv2": len(mesh.vertex_uv2) > 0 if mesh.vertex_uv2 is not None else False,
    }


# GLM-compatible transformation functions
# These provide a GLM-style API for matrix and vector operations


@overload
def translate(v: Vector3, /) -> Matrix4: ...
@overload
def translate(v: Any, /) -> Any: ...
def translate(v: Any, /) -> Any:
    """Create a translation matrix.

    Args:
    ----
        v: Translation vector (Vector3).

    Returns:
    -------
        Matrix4: Translation matrix.

    Note:
    ----
        Matrix4._data is row-major: _data[row][col].
        Translation goes in column 3 (the last column) of rows 0-2.
    """
    result = Matrix4(1.0)  # Start with identity matrix
    result._data[0][3] = v.x
    result._data[1][3] = v.y
    result._data[2][3] = v.z
    return result


@overload
def rotate(m: Matrix4, angle: float, axis: Vector3, /) -> Matrix4: ...
@overload
def rotate(m: Any, angle: float, axis: Any, /) -> Any: ...
def rotate(m: Any, angle: float, axis: Any, /) -> Any:
    """Rotate a matrix by angle (radians) around axis.

    Args:
    ----
        m: Matrix to rotate.
        angle: Rotation angle in radians.
        axis: Rotation axis (Vector3).

    Returns:
    -------
        Matrix4: Rotated matrix.
    """
    # Normalize axis
    axis_length = math.sqrt(axis.x**2 + axis.y**2 + axis.z**2)
    if axis_length == 0:
        return Matrix4(m)

    x = axis.x / axis_length
    y = axis.y / axis_length
    z = axis.z / axis_length

    c = math.cos(angle)
    s = math.sin(angle)
    t = 1 - c

    # Rotation matrix (row-major: _data[row][col])
    rot = Matrix4(0.0)
    rot._data[0][0] = t * x * x + c
    rot._data[0][1] = t * x * y - s * z
    rot._data[0][2] = t * x * z + s * y
    rot._data[1][0] = t * x * y + s * z
    rot._data[1][1] = t * y * y + c
    rot._data[1][2] = t * y * z - s * x
    rot._data[2][0] = t * x * z - s * y
    rot._data[2][1] = t * y * z + s * x
    rot._data[2][2] = t * z * z + c
    rot._data[3][3] = 1.0  # Must be 1 for a proper affine transform

    return m * rot


@overload
def mat4_cast(x: Vector4, /) -> Matrix4: ...
@overload
def mat4_cast(x: Any, /) -> Any: ...
def mat4_cast(x: Any, /) -> Any:
    """Convert a quaternion (Vector4) to a 4x4 rotation matrix.

    Args:
    ----
        x: Quaternion as Vector4 (w, x, y, z order expected, but Vector4 is x, y, z, w).
           For quaternions, typically Vector4(w, x, y, z) is used.

    Returns:
    -------
        Matrix4: Rotation matrix.

    Note:
    ----
        Vector4 stores (x, y, z, w) but quaternions are typically (w, x, y, z).
        This function assumes Vector4 quaternion format is (x, y, z, w).
    """
    # Row-major rotation matrix from quaternion: _data[row][col]
    result = Matrix4(0.0)

    # Vector4 is (x, y, z, w), quaternion is (w, x, y, z)
    qw, qx, qy, qz = x.w, x.x, x.y, x.z

    result._data[0][0] = 1 - 2 * qy * qy - 2 * qz * qz
    result._data[0][1] = 2 * qx * qy - 2 * qz * qw
    result._data[0][2] = 2 * qx * qz + 2 * qy * qw

    result._data[1][0] = 2 * qx * qy + 2 * qz * qw
    result._data[1][1] = 1 - 2 * qx * qx - 2 * qz * qz
    result._data[1][2] = 2 * qy * qz - 2 * qx * qw

    result._data[2][0] = 2 * qx * qz - 2 * qy * qw
    result._data[2][1] = 2 * qy * qz + 2 * qx * qw
    result._data[2][2] = 1 - 2 * qx * qx - 2 * qy * qy

    result._data[3][3] = 1.0  # Proper affine transform

    return result


@overload
def inverse(m: Matrix4, /) -> Matrix4: ...
@overload
def inverse(m: Any, /) -> Any: ...
def inverse(m: Any, /) -> Any:
    """Calculate the inverse of a matrix.

    Args:
    ----
        m: Matrix to invert.

    Returns:
    -------
        Matrix4: Inverse matrix, or identity if singular.
    """
    # Simple 4x4 matrix inversion using Gaussian elimination
    # For numerical stability, we use a basic implementation
    # In production, numpy.linalg.inv would be preferred but we avoid dependencies

    try:
        # Create augmented matrix [M|I]
        aug = [[0.0] * 8 for _ in range(4)]
        for i in range(4):
            for j in range(4):
                aug[i][j] = m._data[i][j]
            aug[i][i + 4] = 1.0

        # Gaussian elimination with partial pivoting
        for col in range(4):
            # Find pivot
            max_row = col
            for row in range(col + 1, 4):
                if abs(aug[row][col]) > abs(aug[max_row][col]):
                    max_row = row

            # Swap rows
            aug[col], aug[max_row] = aug[max_row], aug[col]

            # Check for singular matrix
            if abs(aug[col][col]) < 1e-10:
                return Matrix4(1.0)  # Return identity if singular

            # Normalize pivot row
            pivot = aug[col][col]
            for j in range(8):
                aug[col][j] /= pivot

            # Eliminate column
            for row in range(4):
                if row != col:
                    factor = aug[row][col]
                    for j in range(8):
                        aug[row][j] -= factor * aug[col][j]

        # Extract inverse
        result = Matrix4(0.0)
        for i in range(4):
            for j in range(4):
                result._data[i][j] = aug[i][j + 4]

        return result
    except Exception:
        # Return identity matrix if inversion fails
        return Matrix4(1.0)


def perspective(fovy: float, aspect: float, near: float, far: float, /) -> Matrix4:
    """Create a perspective projection matrix.

    Args:
    ----
        fovy: Field of view in degrees.
        aspect: Aspect ratio (width/height).
        near: Near clipping plane.
        far: Far clipping plane.

    Returns:
    -------
        Matrix4: Perspective projection matrix (row-major: _data[row][col]).

    Note:
    ----
        Standard OpenGL perspective matrix (row-major):
        Row 0: [f/a,  0,    0,                    0              ]
        Row 1: [0,    f,    0,                    0              ]
        Row 2: [0,    0,    -(f+n)/(f-n),         -2*f*n/(f-n)   ]
        Row 3: [0,    0,    -1,                   0              ]
        where f = 1/tan(fovy/2), a = aspect.
    """
    result = Matrix4(0.0)

    fov_rad = math.radians(fovy)
    tan_half_fov = math.tan(fov_rad / 2.0)

    result._data[0][0] = 1.0 / (aspect * tan_half_fov)
    result._data[1][1] = 1.0 / tan_half_fov
    result._data[2][2] = -(far + near) / (far - near)
    result._data[2][3] = -(2.0 * far * near) / (far - near)
    result._data[3][2] = -1.0

    return result


@overload
def normalize(x: Vector3, /) -> Vector3: ...
@overload
def normalize(x: Any, /) -> Any: ...
def normalize(x: Any, /) -> Any:
    """Normalize a vector.

    Args:
    ----
        x: Vector to normalize (Vector3).

    Returns:
    -------
        Vector3: Normalized vector.
    """
    v = x
    length_val = math.sqrt(v.x**2 + v.y**2 + v.z**2)
    if length_val == 0:
        return Vector3(0, 0, 0)
    return Vector3(v.x / length_val, v.y / length_val, v.z / length_val)


@overload
def cross(x: Vector3, y: Vector3, /) -> Vector3: ...
@overload
def cross(x: Any, y: Any, /) -> Any: ...
def cross(x: Any, y: Any, /) -> Any:
    """Calculate the cross product of two vectors.

    Args:
    ----
        x: First vector (Vector3).
        y: Second vector (Vector3).

    Returns:
    -------
        Vector3: Cross product.
    """
    if isinstance(x, Vector3) and isinstance(y, Vector3):
        return Vector3(
            x.y * y.z - x.z * y.y,
            x.z * y.x - x.x * y.z,
            x.x * y.y - x.y * y.x,
        )
    raise TypeError(f"Unsupported types for cross: {x.__class__.__name__}, {y.__class__.__name__}")


@overload
def decompose(modelMatrix: Matrix4, scale: Vector3, orientation: Vector4, translation: Vector3, skew: Vector3, perspective_vec: Vector4, /) -> bool: ...
@overload
def decompose(modelMatrix: Any, scale: Any, orientation: Any, translation: Any, skew: Any, perspective_vec: Any, /) -> bool: ...
def decompose(
    modelMatrix: Any,
    scale: Any,
    orientation: Any,
    translation: Any,
    skew: Any,
    perspective_vec: Any,
    /,
) -> bool:
    """Decompose a transformation matrix into its components.

    Args:
    ----
        modelMatrix: The transformation matrix to decompose.
        scale: Output scale vector (Vector3).
        orientation: Output rotation quaternion (Vector4, as x, y, z, w).
        translation: Output translation vector (Vector3).
        skew: Output skew vector (Vector3, not implemented).
        perspective_vec: Output perspective vector (Vector4, not implemented).

    Returns:
    -------
        bool: True if decomposition was successful.
    """
    if (
        not isinstance(modelMatrix, Matrix4)
        or not isinstance(scale, Vector3)
        or not isinstance(orientation, Vector4)
        or not isinstance(translation, Vector3)
        or not isinstance(skew, Vector3)
        or not isinstance(perspective_vec, Vector4)
    ):
        raise TypeError("decompose requires Matrix4, Vector3, Vector4, Vector3, Vector3, Vector4 arguments")
    m = modelMatrix._data

    # Extract translation (row-major: translation is in column 3, rows 0-2)
    translation.x = float(m[0][3])
    translation.y = float(m[1][3])
    translation.z = float(m[2][3])

    # Extract scale
    scale_x = math.sqrt(m[0][0] ** 2 + m[0][1] ** 2 + m[0][2] ** 2)
    scale_y = math.sqrt(m[1][0] ** 2 + m[1][1] ** 2 + m[1][2] ** 2)
    scale_z = math.sqrt(m[2][0] ** 2 + m[2][1] ** 2 + m[2][2] ** 2)

    scale.x = scale_x
    scale.y = scale_y
    scale.z = scale_z

    # Normalize matrix to extract rotation
    if scale_x == 0 or scale_y == 0 or scale_z == 0:
        orientation.x = 0.0
        orientation.y = 0.0
        orientation.z = 0.0
        orientation.w = 1.0
        return True

    rot_matrix = [
        [m[0][0] / scale_x, m[0][1] / scale_x, m[0][2] / scale_x],
        [m[1][0] / scale_y, m[1][1] / scale_y, m[1][2] / scale_y],
        [m[2][0] / scale_z, m[2][1] / scale_z, m[2][2] / scale_z],
    ]

    # Convert rotation matrix to quaternion
    trace = rot_matrix[0][0] + rot_matrix[1][1] + rot_matrix[2][2]

    if trace > 0:
        s = math.sqrt(trace + 1.0) * 2
        orientation.w = 0.25 * s
        orientation.x = (rot_matrix[2][1] - rot_matrix[1][2]) / s
        orientation.y = (rot_matrix[0][2] - rot_matrix[2][0]) / s
        orientation.z = (rot_matrix[1][0] - rot_matrix[0][1]) / s
    elif rot_matrix[0][0] > rot_matrix[1][1] and rot_matrix[0][0] > rot_matrix[2][2]:
        s = math.sqrt(1.0 + rot_matrix[0][0] - rot_matrix[1][1] - rot_matrix[2][2]) * 2
        orientation.w = (rot_matrix[2][1] - rot_matrix[1][2]) / s
        orientation.x = 0.25 * s
        orientation.y = (rot_matrix[0][1] + rot_matrix[1][0]) / s
        orientation.z = (rot_matrix[0][2] + rot_matrix[2][0]) / s
    elif rot_matrix[1][1] > rot_matrix[2][2]:
        s = math.sqrt(1.0 + rot_matrix[1][1] - rot_matrix[0][0] - rot_matrix[2][2]) * 2
        orientation.w = (rot_matrix[0][2] - rot_matrix[2][0]) / s
        orientation.x = (rot_matrix[0][1] + rot_matrix[1][0]) / s
        orientation.y = 0.25 * s
        orientation.z = (rot_matrix[1][2] + rot_matrix[2][1]) / s
    else:
        s = math.sqrt(1.0 + rot_matrix[2][2] - rot_matrix[0][0] - rot_matrix[1][1]) * 2
        orientation.w = (rot_matrix[1][0] - rot_matrix[0][1]) / s
        orientation.x = (rot_matrix[0][2] + rot_matrix[2][0]) / s
        orientation.y = (rot_matrix[1][2] + rot_matrix[2][1]) / s
        orientation.z = 0.25 * s

    return True


@overload
def eulerAngles(x: Vector4, /) -> Vector3: ...
@overload
def eulerAngles(x: Any, /) -> Any: ...
def eulerAngles(x: Any, /) -> Any:
    """Convert a quaternion (Vector4) to Euler angles (in radians).

    Args:
    ----
        x: Quaternion as Vector4 (x, y, z, w format).

    Returns:
    -------
        Vector3: Euler angles (roll, pitch, yaw) in radians.
    """
    q = x
    # Roll (x-axis rotation)
    sinr_cosp = 2 * (q.w * q.x + q.y * q.z)
    cosr_cosp = 1 - 2 * (q.x * q.x + q.y * q.y)
    roll = math.atan2(sinr_cosp, cosr_cosp)

    # Pitch (y-axis rotation)
    sinp = 2 * (q.w * q.y - q.z * q.x)
    if abs(sinp) >= 1:
        pitch = math.copysign(math.pi / 2, sinp)
    else:
        pitch = math.asin(sinp)

    # Yaw (z-axis rotation)
    siny_cosp = 2 * (q.w * q.z + q.x * q.y)
    cosy_cosp = 1 - 2 * (q.y * q.y + q.z * q.z)
    yaw = math.atan2(siny_cosp, cosy_cosp)

    return Vector3(roll, pitch, yaw)


@overload
def value_ptr(x: Matrix4, /) -> ndarray: ...
@overload
def value_ptr(x: Vector3, /) -> ndarray: ...
@overload
def value_ptr(x: Vector4, /) -> ndarray: ...
@overload
def value_ptr(x: Any, /) -> Any: ...
def value_ptr(x: Any, /) -> Any:
    """Get a pointer to the underlying data (returns flattened numpy array).

    This function requires numpy and is primarily for OpenGL interop.

    Args:
    ----
        x: Matrix or vector to get pointer from.

    Returns:
    -------
        np.ndarray: Flattened array of the data.

    Raises:
    ------
        ImportError: If numpy is not available.
    """
    try:
        import numpy as np
    except ImportError as e:
        raise ImportError("value_ptr requires numpy to be installed") from e

    if isinstance(x, Matrix4):
        # PyKotor Matrix4 - has _data attribute
        # OpenGL expects column-major order
        # Convert row-major to column-major
        col_major = [[x._data[i][j] for i in range(4)] for j in range(4)]
        flat = [col_major[i][j] for i in range(4) for j in range(4)]
        return np.ascontiguousarray(np.array(flat, dtype=np.float32))
    if x.__class__.__name__ == "mat4x4":
        # GLM mat4x4 - already in column-major order
        # GLM matrices are indexable as mat[col][row]
        flat = [x[i][j] for i in range(4) for j in range(4)]
        return np.ascontiguousarray(np.array(flat, dtype=np.float32))
    if isinstance(x, Vector3):
        return np.ascontiguousarray(np.array([x.x, x.y, x.z], dtype=np.float32))
    if isinstance(x, Vector4):
        return np.ascontiguousarray(np.array([x.x, x.y, x.z, x.w], dtype=np.float32))
    raise TypeError(f"value_ptr requires Matrix4, Vector3, or Vector4, got {x.__class__.__name__}")


@overload
def unProject(obj: Vector3, model: Matrix4, proj: Matrix4, viewport: Vector4, /) -> Vector3: ...
@overload
def unProject(obj: Any, model: Any, proj: Any, viewport: Any, /) -> Any: ...
def unProject(obj: Any, model: Any, proj: Any, viewport: Any, /) -> Any:
    """Unproject a window coordinate to world coordinates.

    Args:
    ----
        obj: Window coordinates (Vector3, where z is depth).
        model: Model-view matrix (Matrix4).
        proj: Projection matrix (Matrix4).
        viewport: Viewport as Vector4 (x, y, width, height).

    Returns:
    -------
        Vector3: World coordinates.
    """
    if not isinstance(obj, Vector3) or not isinstance(model, Matrix4) or not isinstance(proj, Matrix4) or not isinstance(viewport, Vector4):
        raise TypeError(
            f"unProject requires Vector3, Matrix4, Matrix4, Vector4, got {obj.__class__.__name__}, {model.__class__.__name__}, {proj.__class__.__name__}, {viewport.__class__.__name__}"
        )
    win = obj
    # Compute the inverse of model * projection
    m: Matrix4 = proj * model  # type: ignore[assignment]
    inv_m: Matrix4 = inverse(m)

    # Normalize window coordinates to NDC [-1, 1]
    ndc = Vector4(
        (win.x - viewport.x) / viewport.z * 2.0 - 1.0,
        (win.y - viewport.y) / viewport.w * 2.0 - 1.0,
        2.0 * win.z - 1.0,
        1.0,
    )

    # Transform NDC to world coordinates
    world: Vector4 = inv_m * ndc  # type: ignore[assignment]

    # Perspective divide
    if world.w != 0:
        world.x /= world.w
        world.y /= world.w
        world.z /= world.w

    return Vector3(world.x, world.y, world.z)


@overload
def length(x: Vector3, /) -> float: ...
@overload
def length(x: Any, /) -> Any: ...
def length(x: Any, /) -> Any:
    """Calculate the length (magnitude) of a vector.

    Args:
    ----
        x: Vector (Vector3).

    Returns:
    -------
        float: Vector length.
    """
    return math.sqrt(x.x**2 + x.y**2 + x.z**2)



"""This module handles classes relating to editing BIF/BZF files.

BIF (Bioware Index File) files are archive containers that store the bulk of game resources.
They work in tandem with KEY files which provide the filename-to-resource mappings. BIF files
contain only resource IDs, types, and data - the actual filenames (ResRefs) are stored in the
KEY file. BZF files are LZMA-compressed BIF files used in some game distributions.

References:
----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        BIF file format specification
        Binary Format:
        -------------
        Header (20 bytes):
        Offset | Size | Type   | Description
        -------|------|--------|-------------
        0x00   | 4    | char[] | File Type ("BIFF" or "BZF ")
        0x04   | 4    | char[] | File Version ("V1  " for BIF, "V1.0" for BZF)
        0x08   | 4    | uint32 | Variable Resource Count
        0x0C   | 4    | uint32 | Fixed Resource Count (unused in KotOR, always 0)
        0x10   | 4    | uint32 | Offset to Variable Resource Table
        Variable Resource Entry (16 bytes each):
        Offset | Size | Type   | Description
        -------|------|--------|-------------
        0x00   | 4    | uint32 | Resource ID (matches KEY file entry)
        0x04   | 4    | uint32 | Offset to resource data
        0x08   | 4    | uint32 | File Size (uncompressed)
        0x0C   | 4    | uint32 | Resource Type
        Fixed Resource Entry (unused in KotOR, 20 bytes each if present):
        Similar to Variable but with additional Part Number field
        Resource Data:
        Raw binary data for each resource at specified offsets

BZF Compression:
---------------
    BZF files use LZMA compression on the entire BIF file after the 8-byte header.
    The BZF header contains: "BZF " + "V1.0", followed by LZMA-compressed BIF data.
    Decompression reveals a standard BIF structure.
"""

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, ClassVar

from pykotor.common.misc import ResRef
from pykotor.extract.file import ResourceIdentifier
from pykotor.resource.bioware_archive import ArchiveResource, BiowareArchive
from pykotor.resource.formats.key import KeyEntry
from pykotor.tools.misc import is_bif_file, is_bzf_file

if TYPE_CHECKING:
    import os

    from pykotor.resource.formats.key import KEY, BifEntry
    from pykotor.resource.type import ResourceType


class BIFType(Enum):
    """The type of BIF file based on file header signature.

    BIF files can be either uncompressed (BIFF) or LZMA-compressed (BZF).
    The file type is determined by the first 4 bytes of the file header.

    References:
    ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET/Formats/KotorBIF/BIFBinaryStructure.cs:36
        https://github.com/th3w1zard1/KotOR_IO/tree/master/KotOR_IO/File


    """

    BIF = "BIFF"  # Regular uncompressed BIF file
    BZF = "BZF "  # LZMA-compressed BIF file (used in some distributions)

    @classmethod
    def from_extension(
        cls,
        ext_or_filepath: os.PathLike | str,
    ) -> BIFType:
        """Get BIF type from file extension."""
        if is_bif_file(ext_or_filepath):
            return cls.BIF
        if is_bzf_file(ext_or_filepath):
            return cls.BZF
        msg = f"'{ext_or_filepath}' is not a valid BZF/BIF file extension."
        raise ValueError(msg)


class BIFResource(ArchiveResource):
    """A single resource entry stored in a BIF/BZF file.

    BIF resources contain only the resource data, type, and ID. The actual filename (ResRef)
    is stored in the KEY file and matched via the resource ID. Each resource has a unique ID
    within the BIF that corresponds to entries in the KEY file's resource table.

    References:
    ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET/Formats/KotorBIF/BIFBinaryStructure.cs:51-65
        https://github.com/th3w1zard1/KotOR_IO/tree/master/KotOR_IO/File



    Attributes:
    ----------
        resname_key_index: Composite resource ID that matches KEY file entries (same bit layout as KeyEntry.resource_id)
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/BIFBinaryStructure.cs:53 (ResourceID property)
            Reference: https://github.com/th3w1zard1/KotOR_IO/tree/master/BIF.cs:203 (ID field)
            Bits 31-20: BIF index in KEY file table. Bits 19-0: resource index within this BIF.
            Used to match resources between BIF and KEY files.

        _offset: Byte offset to resource data within BIF file
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/BIFBinaryStructure.cs:54 (Offset property)
            Reference: https://github.com/th3w1zard1/KotOR_IO/tree/master/BIF.cs:204 (Offset field)
            Points to start of raw resource data in file
            Offsets are absolute from beginning of file

        _packed_size: Size of compressed data (BZF only, unused in regular BIF)
            BZF-specific field for compressed resource size
            Not present in standard BIF format
            Typically equals uncompressed size for BIF files
    """

    def __init__(
        self,
        resref: ResRef,
        restype: ResourceType,
        data: bytes,
        resname_key_index: int | None = None,
        size: int | None = None,
    ):
        super().__init__(resref=resref, restype=restype, data=data, size=size)

        # https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET/Formats/KotorBIF/BIFBinaryStructure.cs:53
        # https://github.com/th3w1zard1/KotOR_IO/tree/master/KotOR_IO/File Formats/BIF.cs:203
        # Resource ID (matches KEY file, unique within BIF)
        self.resname_key_index: int = 0 if resname_key_index is None else resname_key_index

        # https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET/Formats/KotorBIF/BIFBinaryStructure.cs:54
        # https://github.com/th3w1zard1/KotOR_IO/tree/master/KotOR_IO/File Formats/BIF.cs:204
        # Byte offset to resource data in file
        self._offset: int = 0  # Offset in BIF file

        # BZF-specific: Size of compressed data
        self._packed_size: int = 0  # Size of compressed data (BZF only)

    @property
    def offset(self) -> int:
        """Get resource offset."""
        return self._offset

    @offset.setter
    def offset(self, value: int) -> None:
        """Set resource offset."""
        self._offset = value

    @property
    def packed_size(self) -> int:
        """Get compressed data size (BZF only)."""
        return self._packed_size

    @packed_size.setter
    def packed_size(self, value: int) -> None:
        """Set compressed data size (BZF only)."""
        self._packed_size = value

    def __eq__(
        self,
        other: object,
    ) -> bool:
        """Compare two resources."""
        if not isinstance(other, BIFResource):
            return NotImplemented  # type: ignore[no-any-return]
        return self.resname_key_index == other.resname_key_index and self.restype == other.restype and self.size == other.size

    def __hash__(self) -> int:
        """Hash resource."""
        return hash((self.resname_key_index, self.restype, self.size))

    def __str__(self) -> str:
        """Get string representation."""
        return f"{self.resref}:{self.restype.name}[{self.size}b]"


class BIF(BiowareArchive):
    """Represents a BIF/BZF file in the Aurora engine.

    BIF (Binary Index Format) files are the primary data containers for KotOR game resources.
    They store thousands of game assets (models, textures, scripts, etc.) in a single file.
    BIF files work in conjunction with KEY files: the BIF contains the data and resource IDs,
    while the KEY file maps filenames (ResRefs) to resource IDs and BIF locations.

    References:
    ----------
        Original BioWare engine binaries (from swkotor.exe, swkotor2.exe)
        Original BioWare engine binaries
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET/Formats/KotorBIF/BIFBinaryStructure.cs:15-32
        https://github.com/th3w1zard1/KotOR_IO/tree/master/KotOR_IO/File
        https://github.com/th3w1zard1/KotOR.js/tree/master/src/resource/BIFObject.ts:11-152



    Attributes:
    ----------
        HEADER_SIZE: Size of BIF header in bytes (20 bytes)
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/BIFBinaryStructure.cs:41-47 (header fields)
            Reference: https://github.com/th3w1zard1/KotOR_IO/tree/master/BIF.cs:46-51 (header parsing)
            Fixed size across all BIF versions

        VAR_ENTRY_SIZE: Size of each variable resource entry (16 bytes)
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/BIFBinaryStructure.cs:58-64 (VariableResource reading)
            Reference: https://github.com/th3w1zard1/KotOR_IO/tree/master/BIF.cs:56 (entry reading loop)
            Each entry: ID(4) + Offset(4) + Size(4) + Type(4)

        FIX_ENTRY_SIZE: Size of fixed resource entry (16-20 bytes, unused in KotOR)
            Reference: https://github.com/th3w1zard1/KotOR_IO/tree/master/BIF.cs:63 (FixedResourceEntry struct)
            Fixed resources not used in KotOR games (always 0 count)

        FILE_VERSION: BIF file format version ("V1  ")
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/BIFBinaryStructure.cs:44 (FileVersion)
            Reference: https://github.com/th3w1zard1/KotOR_IO/tree/master/BIF.cs:48 (Version field)

        bif_type: Whether this is regular BIF or compressed BZF
            Determines compression handling during load/save

        _resources: List of all resources in this BIF
            Reference: https://github.com/th3w1zard1/Kotor.NET/tree/master/BIFBinaryStructure.cs:18 (Resources list)
            Reference: https://github.com/th3w1zard1/KotOR_IO/tree/master/BIF.cs:96 (VariableResourceTable)
            Ordered list maintained for indexing and iteration

        _resource_dict: Fast lookup by ResRef and ResourceType
            PyKotor-specific optimization for O(1) resource lookup
            Built from KEY file data or user assignment

        _id_lookup: Fast lookup by resource ID
            Reference: Similar to KEY file's resource_by_id mapping
            Maps resource ID to BIFResource for KEY-BIF coordination

        _modified: Tracks if BIF has been modified since loading
            Used to determine if file needs to be saved
    """

    HEADER_SIZE: ClassVar[int] = 20  # Fixed header size
    VAR_ENTRY_SIZE: ClassVar[int] = 16  # Size of each variable resource entry
    FIX_ENTRY_SIZE: ClassVar[int] = 16  # Size of each fixed resource entry

    FILE_VERSION: ClassVar[str] = "V1  "

    def __init__(
        self,
        bif_type: BIFType = BIFType.BIF,
    ):
        super().__init__()

        # File type (BIF vs BZF determines compression)
        self.bif_type: BIFType = bif_type

        # https://github.com/th3w1zard1/Kotor.NET/tree/master/Kotor.NET/Formats/KotorBIF/BIFBinaryStructure.cs:18
        # https://github.com/th3w1zard1/KotOR_IO/tree/master/KotOR_IO/File Formats/BIF.cs:96
        # List of all resources in file (ordered)
        self._resources: list[BIFResource] = []

        # PyKotor optimization: ResRef+Type -> Resource lookup (O(1) access)
        self._resource_dict: dict[ResourceIdentifier, BIFResource] = {}

        # PyKotor optimization: Resource ID -> Resource lookup (for KEY coordination)
        self._id_lookup: dict[int, BIFResource] = {}

        # Modification tracking flag
        self._modified: bool = False

    @property
    def resources(self) -> list[BIFResource]:
        """Get list of resources."""
        return self._resources

    @property
    def var_count(self) -> int:
        """Number of variable resources."""
        return len(self._resources)

    @property
    def fixed_count(self) -> int:
        """Number of fixed resources."""
        return 0  # Currently no fixed resources supported

    def get_resource_offset(
        self,
        resource: ArchiveResource,
    ) -> int:
        """Get offset and size for a resource.

        Required by BiowareArchive.
        """
        if not isinstance(resource, BIFResource):
            msg = "Resource is not a BIFResource: cannot get offset"
            raise TypeError(msg)
        try:
            if self.bif_type == BIFType.BZF:
                return resource.offset
        except ValueError as e:
            msg = "Resource not found in BIF"
            raise ValueError(msg) from e
        return resource.offset

    def set_data(
        self,
        resref: ResRef,
        restype: ResourceType,
        data: bytes,
        res_id: int | None = None,
    ) -> BIFResource:
        """Create and add a new resource."""
        resource = BIFResource(ResRef(resref), restype, data, res_id)
        self._resources.append(resource)
        self._resource_dict[ResourceIdentifier(str(resref), restype)] = resource
        self._id_lookup[resource.resname_key_index] = resource
        self._modified = True
        return resource

    def remove_resource(
        self,
        resource: BIFResource,
    ) -> None:
        """Remove a resource from the BIF."""
        try:
            self._resources.remove(resource)
            del self._resource_dict[ResourceIdentifier(str(resource.resref), resource.restype)]
            del self._id_lookup[resource.resname_key_index]
            self._modified = True
        except (ValueError, KeyError) as e:
            msg = f"Resource '{resource!r}' not found in BIF"
            raise ValueError(msg) from e

    def reorder_resources(
        self,
        new_order: list[BIFResource],
    ) -> None:
        """Reorder resources in the BIF."""
        if set(new_order) != set(self._resources):
            msg = "New order must contain exactly the same resources"
            raise ValueError(msg)
        self._resources = new_order.copy()
        self.build_lookup_tables()
        self._modified = True

    def get_resource_by_id(
        self,
        resource_id: int,
    ) -> BIFResource | None:
        """Get resource by ID."""
        return self._id_lookup.get(resource_id)

    def get_resources_by_type(
        self,
        restype: ResourceType,
    ) -> list[BIFResource]:
        """Get all resources of a specific type."""
        return [res for res in self._resources if res.restype == restype]

    def try_get_resource(
        self,
        resref: str | ResRef,
        restype: ResourceType,
    ) -> tuple[bool, BIFResource | None]:
        """Try to get resource by ResRef and type."""
        if isinstance(resref, ResRef):
            resref = str(resref)
        resource: BIFResource | None = self._resource_dict.get(ResourceIdentifier(resref.lower(), restype))
        return (resource is not None, resource)

    def build_lookup_tables(self) -> None:
        """Build internal lookup tables for fast resource access."""
        self._resource_dict.clear()
        self._id_lookup.clear()
        for resource in self._resources:
            self._resource_dict[ResourceIdentifier(str(resource.resref), resource.restype)] = resource
            self._id_lookup[resource.resname_key_index] = resource

    @property
    def is_compressed(self) -> bool:
        """Check if this is a compressed BIF (BZF)."""
        return self.bif_type is BIFType.BZF

    @property
    def is_modified(self) -> bool:
        """Whether the BIF has been modified since loading."""
        return self._modified

    def generate_key(
        self,
        filename: str,
        existing_key: KEY | None = None,
    ) -> KEY:
        """Generate a KEY file from this BIF's contents.

        Args:
            filename: The filename to use for this BIF in the KEY.

        Returns:
            A new KEY object containing entries for all resources in this BIF.
        """
        from pykotor.resource.formats.key.key_data import KEY, KeyEntry

        key: KEY = KEY() if existing_key is None else existing_key

        # Add this BIF file
        bif_idx: BifEntry = key.add_bif(filename)

        # Add all resources to the KEY
        for resource in self._resources:
            key.key_entries.append(
                KeyEntry(
                    resref=resource.resref,
                    restype=resource.restype,
                    resid=resource.resname_key_index,
                )
            )

        return key

    def validate_with_key(
        self,
        key: KEY,
        bif_idx: int,
    ) -> list[str]:
        """Validate this BIF's contents against a KEY file.

        Args:
            key: The KEY file to validate against.
            bif_idx: The index of this BIF in the KEY file.

        Returns:
            A list of validation errors, empty if validation passed.
        """
        errors: list[str] = []

        # Get all KEY resources that should be in this BIF
        key_resources: list[KeyEntry] = [r for r in key.key_entries if r.bif_index == bif_idx]

        # Check all KEY resources exist in BIF
        key_res: KeyEntry | None
        for key_res in key_resources:
            bif_res: BIFResource | None = self.get_resource_by_id(key_res.resname_key_index)
            if bif_res is None:
                errors.append(f"Resource {key_res.resref}:{key_res.restype} from KEY not found in BIF")
                continue

            if bif_res.restype != key_res.restype:
                errors.append(f"Resource {key_res.resref} type mismatch: KEY={key_res.restype}, BIF={bif_res.restype}")

        # Check all BIF resources exist in KEY
        for bif_res in self._resources:
            key_res = next((r for r in key_resources if r.resname_key_index == bif_res.resname_key_index), None)
            if key_res is None:
                errors.append(f"Resource ID {bif_res.resname_key_index} from BIF not found in KEY")

        return errors

    def synchronize_with_key(
        self,
        key: KEY,
        bif_idx: int,
    ) -> None:
        """Synchronize this BIF's resource names with a KEY file.

        Args:
            key: The KEY file to synchronize with.
            bif_idx: The index of this BIF in the KEY file.
        """
        # Get all KEY resources for this BIF
        key_resources: dict[int, KeyEntry] = {r.resname_key_index: r for r in key.key_entries if r.bif_index == bif_idx}

        # Update BIF resource names from KEY
        for resource in self._resources:
            key_res: KeyEntry | None = key_resources.get(resource.resname_key_index)
            if key_res is None:
                continue
            resource.resref = key_res.resref

        self.build_lookup_tables()

    def create_key(
        self,
        filename: str,
    ) -> KEY:
        """Create a KEY file from this BIF's contents.

        Args:
            filename: The filename to use for this BIF in the KEY.

        Returns:
            A new KEY object containing entries for all resources in this BIF.
        """
        from pykotor.resource.formats.key.key_data import KEY

        key = KEY()
        key.add_bif(filename)

        # Add all resources to the KEY
        for resource in self._resources:
            key_res = KeyEntry(resref=resource.resref, restype=resource.restype, resid=resource.resname_key_index)
            key.key_entries.append(key_res)

        return key

    def apply_key(
        self,
        key: KEY,
        bif_idx: int,
    ) -> list[str]:
        """Apply resource names from a KEY file to this BIF.

        Args:
            key: The KEY file to get names from.
            bif_idx: The index of this BIF in the KEY file.

        Returns:
            List of any errors encountered during the process.
        """
        errors: list[str] = []

        # Get all KEY resources for this BIF
        key_resources: dict[int, KeyEntry] = {r.resname_key_index: r for r in key.key_entries if r.bif_index == bif_idx}

        # Update BIF resource names from KEY
        for resource in self._resources:
            if key_res := key_resources.get(resource.resname_key_index):
                if resource.restype != key_res.restype:
                    errors.append(f"Resource type mismatch for ID {resource.resname_key_index}: BIF={resource.restype}, KEY={key_res.restype}")
                resource.resref = key_res.resref
            else:
                errors.append(f"Resource ID {resource.resname_key_index} from BIF not found in KEY")

        # Check for KEY resources not in BIF
        bif_ids: set[int] = {r.resname_key_index for r in self._resources}
        errors.extend(
            [
                f"Resource {key_res.resref}:{key_res.restype} from KEY not found in BIF (ID: {key_res.resname_key_index})"
                for key_res in key_resources.values()
                if key_res.resname_key_index not in bif_ids
            ]
        )  # noqa: E501

        # Rebuild lookup tables with new names
        self.build_lookup_tables()

        return errors

    def reorganize_by_type(self) -> None:
        """Reorganize resources by grouping them by type.

        This can improve access performance when reading resources of the same type.
        """
        self._resources.sort(key=lambda r: (r.restype.type_id, r.resname_key_index))
        self._modified = True
        self.build_lookup_tables()



from __future__ import annotations

import uuid

from abc import abstractmethod
from contextlib import suppress
from typing import TYPE_CHECKING, Any, cast

from qtpy.QtCore import QEvent, QObject, QSortFilterProxyModel, Qt
from qtpy.QtGui import QStandardItemModel
from qtpy.QtWidgets import (
    QAbstractSpinBox,
    QApplication,
    QComboBox,
    QDoubleSpinBox,
    QGroupBox,
    QSlider,
    QSpinBox,
    QWidget,
)

from loggerplus import RobustLogger

if TYPE_CHECKING:
    from qtpy.QtCore import QModelIndex
    from qtpy.QtGui import QKeyEvent


class TemplateFilterProxyModel(QSortFilterProxyModel):
    @abstractmethod
    def get_sort_value(self, index: QModelIndex) -> Any: ...


class RobustSortFilterProxyModel(TemplateFilterProxyModel):
    def __init__(
        self,
        parent: QObject | None = None,
    ):
        super().__init__(parent)
        self.setDynamicSortFilter(True)
        self._sort_states: dict[int, int] = {}

    def toggle_sort(self, column: int):
        if column not in self._sort_states:
            self._sort_states[column] = 0
        self._sort_states[column] = (self._sort_states[column] + 1) % 3

        if self._sort_states[column] == 0:
            self.reset_sort()
        elif self._sort_states[column] == 1:
            self.sort(column, Qt.AscendingOrder)
        elif self._sort_states[column] == 2:
            self.sort(column, Qt.DescendingOrder)

    def reset_sort(self):
        self.sort(-1)  # Reset sorting
        self.invalidate()  # Force a refresh
        self._sort_states = {}

    def get_sort_value(self, index: QModelIndex) -> Any:
        """Return the sort value based on the column."""
        srcModel = self.source_model()
        assert isinstance(srcModel, QStandardItemModel)
        return self.source_model().data(index)

    def lessThan(self, left: QModelIndex, right: QModelIndex) -> bool:
        left_data = self.get_sort_value(left)
        right_data = self.get_sort_value(right)

        with suppress(Exception):
            if type(left_data) is type(right_data) and hasattr(left_data, "__lt__"):
                return left_data < right_data
        return super().lessThan(left, right)


class NoScrollEventFilter(QObject):
    def eventFilter(self, obj: QObject, event: QEvent) -> bool:
        if event.type() == QEvent.Type.Wheel and isinstance(obj, QWidget):
            parent_widget = obj.parent()
            self_parent = self.parent()
            while parent_widget is not None and (not isinstance(parent_widget, self_parent.__class__) or self_parent.__class__ == QObject):
                parent_widget = parent_widget.parent()
            if parent_widget:
                QApplication.sendEvent(parent_widget, event)
            return True
        return super().eventFilter(obj, event)

    def setup_filter(
        self,
        include_types: list[type[QWidget]] | None = None,
        parent_widget: QObject | QWidget | None = None,
    ) -> None:
        """Recursively install event filters on all child widgets."""
        if include_types is None:
            include_types = [QComboBox, QSlider, QSpinBox, QGroupBox, QAbstractSpinBox, QDoubleSpinBox]

        parent_widget = self.parent() if parent_widget is None else parent_widget
        if parent_widget is None:
            RobustLogger().warning("NoScrollEventFilter has nothing to do, please provide a widget to process (parent_widget was somehow None here)", stack_info=True)
        for widget in parent_widget.findChildren(QWidget):
            try:
                if not widget.objectName():
                    widget.setObjectName(widget.__class__.__name__ + uuid.uuid4().hex[6:])
            except Exception:  # noqa: BLE001
                RobustLogger().exception(f"Failed to set a temporary object name on {widget}")
            if isinstance(widget, tuple(include_types)):
                # RobustLogger.debug(f"Installing event filter on: {widget.objectName()} (type: {widget.__class__.__name__})")
                widget.installEventFilter(self)
            # else:
            #    RobustLogger.debug(f"Skipping NoScrollEventFilter installation on '{widget.objectName()}' due to instance check {widget.__class__.__name__}.")
            self.setup_filter(include_types, widget)


class HoverEventFilter(QObject):
    def __init__(self, debugKey: Qt.Key | None = None):
        super().__init__()
        self.current_widget: QObject | None = None
        self.debugKey: Qt.Key = Qt.Key_Pause if debugKey is None else debugKey

    def eventFilter(self, obj: QObject, event: QEvent) -> bool:
        if event.type() == QEvent.Type.HoverEnter:
            self.current_widget = obj
        elif event.type() == QEvent.Type.HoverLeave:
            if self.current_widget == obj:
                self.current_widget = None
        elif event.type() == QEvent.Type.KeyPress and cast("QKeyEvent", event).key() == self.debugKey:
            if self.current_widget:
                print(f"Hovered control: {self.current_widget.__class__.__name__} ({self.current_widget.objectName()})")
            else:
                print("No control is currently hovered.")
        return super().eventFilter(obj, event)



"""Game/path heuristics: detect KotOR 1 vs 2 and installation type from directory contents."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from pykotor.common.misc import Game

if TYPE_CHECKING:
    import os


def determine_game(
    path: os.PathLike | str,
) -> Game | None:
    """Determines the game based on files and folders.

    Args:
    ----
        path: Path to game directory

    Returns:
    -------
        Game: Game enum or None

    References:
    ----------
        Based on swkotor.exe ERF structure:
        - See pykotor.resource.formats.erf.erf_data for addresses (K1 + TSL TODO). CExoEncapsulatedFile, AddEncapsulatedContents.
        Original BioWare engine binaries
        Derivations and Other Implementations:
        ----------
        https://github.com/th3w1zard1/HoloPatcher.NET/tree/master/src/HoloPatcher/Utils (Game detection logic)


        Note: File and folder heuristics vary between Steam, GOG, and disc releases

    Processing Logic:
    ----------------
        1. Normalize the path and check for existence of game files
        2. Define checks for each game
        3. Run checks and score games
        4. Return game with highest score or None if scores are equal or all checks fail
    """
    r_path: Path = Path(path)

    def check(x: str) -> bool:
        c_path: Path = r_path.joinpath(x)
        return c_path.exists() is not False

    # Checks for each game
    game1_pc_checks: list[bool] = [
        check("streamwaves"),
        check("swkotor.exe"),
        check("swkotor.ini"),
        check("rims"),
        check("utils"),
        check("32370_install.vdf"),
        check("miles/mssds3d.m3d"),
        check("miles/msssoft.m3d"),
        check("data/party.bif"),
        check("data/player.bif"),
        check("modules/global.mod"),
        check("modules/legal.mod"),
        check("modules/mainmenu.mod"),
    ]

    game1_xbox_checks: list[bool] = [
        check("01_SS_Repair01.ini"),
        check("swpatch.ini"),
        check("dataxbox/_newbif.bif"),
        check("rimsxbox"),
        check("players.erf"),
        check("downloader.xbe"),
        check("rimsxbox/manm28ad_adx.rim"),
        check("rimsxbox/miniglobal.rim"),
        check("rimsxbox/miniglobaldx.rim"),
        check("rimsxbox/STUNT_56a_a.rim"),
        check("rimsxbox/STUNT_56a_adx.rim"),
        check("rimsxbox/STUNT_57_adx.rim"),
        check("rimsxbox/subglobal.rim"),
        check("rimsxbox/subglobaldx.rim"),
        check("rimsxbox/unk_m44ac_adx.rim"),
        check("rimsxbox/M12ab_adx.rim"),
        check("rimsxbox/mainmenu.rim"),
        check("rimsxbox/mainmenudx.rim"),
        check("rimsxbox/manm28ad_adx.rim"),
    ]

    game1_ios_checks: list[bool] = [
        check("override/ios_action_bg.tga"),
        check("override/ios_action_bg2.tga"),
        check("override/ios_action_x.tga"),
        check("override/ios_action_x2.tga"),
        check("override/ios_button_a.tga"),
        check("override/ios_button_x.tga"),
        check("override/ios_button_y.tga"),
        check("override/ios_edit_box.tga"),
        check("override/ios_enemy_plus.tga"),
        check("override/ios_gpad_bg.tga"),
        check("override/ios_gpad_gen.tga"),
        check("override/ios_gpad_gen2.tga"),
        check("override/ios_gpad_help.tga"),
        check("override/ios_gpad_help2.tga"),
        check("override/ios_gpad_map.tga"),
        check("override/ios_gpad_map2.tga"),
        check("override/ios_gpad_save.tga"),
        check("override/ios_gpad_save2.tga"),
        check("override/ios_gpad_solo.tga"),
        check("override/ios_gpad_solo2.tga"),
        check("override/ios_gpad_solox.tga"),
        check("override/ios_gpad_solox2.tga"),
        check("override/ios_gpad_ste.tga"),
        check("override/ios_gpad_ste2.tga"),
        check("override/ios_gpad_ste3.tga"),
        check("override/ios_help.tga"),
        check("override/ios_help2.tga"),
        check("override/ios_help_1.tga"),
        check("KOTOR"),
        check("KOTOR.entitlements"),
        check("kotorios-Info.plist"),
        check("AppIcon29x29.png"),
        check("AppIcon50x50@2x~ipad.png"),
        check("AppIcon50x50~ipad.png"),
    ]

    game1_android_checks: list[bool] = [  # TODO(th3w1zard1): Implement
    ]

    game2_pc_checks: list[bool] = [
        check("streamvoice"),
        check("swkotor2.exe"),
        check("swkotor2.ini"),
        check("LocalVault"),
        check("LocalVault/test.bic"),
        check("LocalVault/testold.bic"),
        check("miles/binkawin.asi"),
        check("miles/mssds3d.flt"),
        check("miles/mssdolby.flt"),
        check("miles/mssogg.asi"),
        check("data/Dialogs.bif"),
    ]

    game2_xbox_checks: list[bool] = [
        check("combat.erf"),
        check("effects.erf"),
        check("footsteps.erf"),
        check("footsteps.rim"),
        check("SWRC"),
        check("weapons.ERF"),
        check("SuperModels/smseta.erf"),
        check("SuperModels/smsetb.erf"),
        check("SuperModels/smsetc.erf"),
        check("SWRC/System/Subtitles_Epilogue.int"),
        check("SWRC/System/Subtitles_YYY_06.int"),
        check("SWRC/System/SWRepublicCommando.int"),
        check("SWRC/System/System.ini"),
        check("SWRC/System/UDebugMenu.u"),
        check("SWRC/System/UnrealEd.int"),
        check("SWRC/System/UnrealEd.u"),
        check("SWRC/System/User.ini"),
        check("SWRC/System/UWeb.int"),
        check("SWRC/System/Window.int"),
        check("SWRC/System/WinDrv.int"),
        check("SWRC/System/Xbox"),
        check("SWRC/System/XboxLive.int"),
        check("SWRC/System/XGame.u"),
        check("SWRC/System/XGameList.int"),
        check("SWRC/System/XGames.int"),
        check("SWRC/System/XInterface.u"),
        check("SWRC/System/XInterfaceMP.u"),
        check("SWRC/System/XMapList.int"),
        check("SWRC/System/XMaps.int"),
        check("SWRC/System/YYY_TitleCard.int"),
        check("SWRC/System/Xbox/Engine.int"),
        check("SWRC/System/Xbox/XboxLive.int"),
        check("SWRC/Textures/GUIContent.utx"),
    ]

    game2_ios_checks: list[bool] = [
        check("override/ios_mfi_deu.tga"),
        check("override/ios_mfi_eng.tga"),
        check("override/ios_mfi_esp.tga"),
        check("override/ios_mfi_fre.tga"),
        check("override/ios_mfi_ita.tga"),
        check("override/ios_self_box_r.tga"),
        check("override/ios_self_expand2.tga"),
        check("override/ipho_forfeit.tga"),
        check("override/ipho_forfeit2.tga"),
        check("override/kotor2logon.tga"),
        check("override/lbl_miscroll_open_f.tga"),
        check("override/lbl_miscroll_open_f2.tga"),
        check("override/ydialog.gui"),
        check("KOTOR II"),
        check("KOTOR2-Icon-20-Apple.png"),
        check("KOTOR2-Icon-29-Apple.png"),
        check("KOTOR2-Icon-40-Apple.png"),
        check("KOTOR2-Icon-58-apple.png"),
        check("KOTOR2-Icon-60-apple.png"),
        check("KOTOR2-Icon-76-apple.png"),
        check("KOTOR2-Icon-80-apple.png"),
        check("KOTOR2_LaunchScreen.storyboardc"),
        check("KOTOR2_LaunchScreen.storyboardc/Info.plist"),
        check("GoogleService-Info.plist"),
    ]

    game2_android_checks: list[bool] = [  # TODO(th3w1zard1): Implement
    ]

    # Determine the game with the most checks passed
    def determine_highest_scoring_game() -> Game | None:
        # Scoring for each game and platform
        scores: dict[Game, int] = {
            Game.K1: sum(game1_pc_checks),
            Game.K2: sum(game2_pc_checks),
            Game.K1_XBOX: sum(game1_xbox_checks),
            Game.K2_XBOX: sum(game2_xbox_checks),
            Game.K1_IOS: sum(game1_ios_checks),
            Game.K2_IOS: sum(game2_ios_checks),
            Game.K1_ANDROID: sum(game1_android_checks),
            Game.K2_ANDROID: sum(game2_android_checks),
        }

        highest_scoring_game: Game | None = None
        highest_score: int = 0

        for game, score in scores.items():
            if score > highest_score:
                highest_score = score
                highest_scoring_game = game

        return highest_scoring_game

    return determine_highest_scoring_game()



"""Reference cache: GFF/NCS/TLK/2DA lookups and installation-wide reference finding."""

from __future__ import annotations

import os
import traceback

from contextlib import suppress
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, Sequence, Union

from loggerplus import RobustLogger
from pykotor.common.language import LocalizedString
from pykotor.common.stream import BinaryReader
from pykotor.extract.file import FileResource  # noqa: PLC0415

# Runtime import for find_tlk_entry_references
from pykotor.extract.installation import SearchLocation  # noqa: PLC0415
from pykotor.extract.twoda import K1Columns2DA, K2Columns2DA  # noqa: PLC0415
from pykotor.resource.formats.gff.gff_auto import read_gff
from pykotor.resource.formats.gff.gff_data import GFFContent, GFFFieldType, GFFList, GFFStruct
from pykotor.resource.formats.ncs.ncs_auto import read_ncs  # noqa: PLC0415
from pykotor.resource.formats.ssf.ssf_auto import read_ssf
from pykotor.resource.formats.ssf.ssf_data import SSF, SSFSound
from pykotor.resource.formats.twoda.twoda_auto import read_2da
from pykotor.resource.type import ResourceType

if TYPE_CHECKING:
    from pathlib import Path

    from pykotor.common.misc import Game
    from pykotor.extract.capsule import Capsule
    from pykotor.extract.file import ResourceIdentifier
    from pykotor.extract.installation import Installation
    from pykotor.resource.formats.gff import GFF
    from pykotor.resource.formats.ncs.ncs_data import NCS
    from pykotor.resource.formats.ssf.ssf_data import SSF
    from pykotor.resource.formats.twoda.twoda_data import TwoDA


# Global cache for scan results to avoid re-scanning unchanged resources
# Key: (filepath, offset, size, mtime) -> Value: list of (strref, location) tuples
_SCAN_RESULTS_CACHE: dict[tuple, list[tuple[int, str]]] = {}


def clear_scan_cache() -> None:
    """Clear the global scan results cache to free memory."""
    _SCAN_RESULTS_CACHE.clear()


# Late import to avoid circular dependency
def _get_2da_columns():
    from pykotor.extract.twoda import K1Columns2DA, K2Columns2DA

    return K1Columns2DA, K2Columns2DA


# Import GFF field mapping from twoda module
def _get_gff_field_to_2da_mapping():
    from pykotor.extract.twoda import TwoDARegistry

    return TwoDARegistry.gff_field_mapping()


# GFF field name to 2DA filename mappings
# Maps GFF field names to the 2DA files they reference
# This is used by TwoDAMemoryReferenceCache and other modules that need to know
# which GFF fields reference which 2DA files
GFF_FIELD_TO_2DA_MAPPING: dict[str, ResourceIdentifier] = _get_gff_field_to_2da_mapping()


# Logging helpers for reference caches
# Log level: 0 = normal, 1 = verbose, 2 = debug
_log_level = 2 if os.environ.get("KOTORDIFF_DEBUG") else (1 if os.environ.get("KOTORDIFF_VERBOSE") else 0)


def _log_debug(msg: str) -> None:
    """Log debug message if debug level is enabled."""
    if _log_level >= 2:  # noqa: PLR2004
        print(f"[DEBUG] {msg}")


def _log_verbose(msg: str) -> None:
    """Log verbose message if verbose level is enabled."""
    if _log_level >= 1:
        print(f"[VERBOSE] {msg}")


# Location dataclasses for StrRef references
@dataclass
class TwoDARefLocation:
    """A reference to a StrRef in a 2DA file."""

    row_index: int
    column_name: str


@dataclass
class SSFRefLocation:
    """A reference to a StrRef in an SSF file."""

    sound: SSF
    sound_index: int


@dataclass
class GFFRefLocation:
    """A reference to a StrRef in a GFF file."""

    field_path: str  # e.g. "FirstName", "Description", "ItemList[0].LocalizedName"


@dataclass
class NCSRefLocation:
    """A reference to a StrRef in an NCS file."""

    byte_offset: int


# Union type for all reference location types
StrRefLocation = Union[TwoDARefLocation, SSFRefLocation, GFFRefLocation, NCSRefLocation]


@dataclass
class StrRefSearchResult:
    """Result of a StrRef reference search containing file resource and typed location objects."""

    resource: FileResource
    locations: Sequence[StrRefLocation]


class StrRefReferenceCache:
    """Cache of StrRef references found during resource scanning.

    Maps StrRef -> list of (resource_identifier, locations) where it's referenced.
    """

    def __init__(self, game: Game):
        """Initialize cache with game-specific 2DA column definitions.

        Args:
            game: Game instance for determining which 2DA columns contain StrRefs
        """
        self.game: Game = game

        # Map: strref -> {resource_identifier: [field_paths]}
        # Using dict for O(1) lookups instead of list for O(n) linear search
        self._cache: dict[int, dict[ResourceIdentifier, list[str]]] = {}

        # Statistics
        self._total_references_found: int = 0
        self._files_with_strrefs: set[str] = set()

        # Get game-specific 2DA column definitions
        K1Columns2DA, K2Columns2DA = _get_2da_columns()
        if game.is_k1():
            self._strref_2da_columns = K1Columns2DA.StrRefs.as_dict()
        elif game.is_k2():
            self._strref_2da_columns = K2Columns2DA.StrRefs.as_dict()
        else:
            self._strref_2da_columns = {}

    def scan_resource(
        self,
        resource: FileResource,
        data: bytes,
    ) -> None:
        """Scan a resource for StrRef references and cache them.

        Args:
            resource: FileResource being scanned
            data: Resource data bytes
        """
        identifier: ResourceIdentifier = resource.identifier()
        restype: ResourceType = resource.restype()
        filename: str = resource.filename().lower()

        try:
            # 2DA files
            if restype == ResourceType.TwoDA and filename in self._strref_2da_columns:
                self._scan_2da(identifier, data, filename)

            # SSF files
            elif restype == ResourceType.SSF:
                self._scan_ssf(identifier, data)

            # NCS files - FIXME: TEMPORARILY DISABLED (will revisit later)
            # elif restype == ResourceType.NCS:
            #     self._scan_ncs(identifier, data)

            # GFF files (only try to parse if restype indicates it's a GFF file)
            elif restype.is_gff():
                try:
                    gff_obj = read_gff(data)
                    self._scan_gff(identifier, gff_obj.root)
                except Exception:  # noqa: BLE001, S110
                    # Failed to parse GFF, skip
                    pass
            # All other file types (textures, sounds, etc.) don't contain StrRefs, skip silently

        except Exception:  # noqa: BLE001, S110
            # Skip files that fail to scan
            pass

    def _scan_2da(
        self,
        identifier: ResourceIdentifier,
        data: bytes,
        filename: str,
    ) -> None:
        """Scan 2DA file for StrRef references."""
        twoda_obj = read_2da(data)
        columns_with_strrefs: set[str] = self._strref_2da_columns[filename]

        for row_idx in range(twoda_obj.get_height()):
            for column_name in tuple(columns_with_strrefs):
                if column_name == ">>##HEADER##<<":
                    continue

                cell = twoda_obj.get_cell(row_idx, column_name)
                if cell and cell.strip().isdigit():
                    strref = int(cell.strip())
                    location = f"row_{row_idx}.{column_name}"
                    _log_debug(f"Found StrRef {strref} in 2DA file '{filename}' at row {row_idx}, column '{column_name}'")
                    self._add_reference(strref, identifier, location)

    def _scan_ssf(
        self,
        identifier: ResourceIdentifier,
        data: bytes,
    ) -> None:
        """Scan SSF file for StrRef references."""
        ssf_obj = read_ssf(data)
        filename: str = f"{identifier.resname}.{identifier.restype.extension}"

        for sound_index in range(28):
            strref: int | None = ssf_obj.get(sound_index)
            if strref is not None and strref != -1:
                location = f"sound_{sound_index}"
                _log_debug(f"Found StrRef {strref} in SSF file '{filename}' at sound slot '{sound_index}'")
                self._add_reference(strref, identifier, location)

    def _scan_ncs(
        self,
        identifier: ResourceIdentifier,
        data: bytes,
    ) -> None:
        """Scan NCS file for StrRef references in CONSTI instructions (optimized).

        FIXME: TEMPORARILY DISABLED - Will revisit NCS scanning later.
        """
        # TEMPORARILY DISABLED - NCS scanning will be revisited later
        return
        # try:
        #     # Validate header - use direct bytes comparison instead of BinaryReader
        #     if len(data) < 13 or data[:4] != b"NCS " or data[4:8] != b"V1.0" or data[8] != 0x42:  # noqa: PLR2004
        #         return
        #
        #     # Extract total_size using struct (faster than BinaryReader)
        #     total_size = struct.unpack(">I", data[9:13])[0]
        #
        #     # Use direct byte indexing instead of BinaryReader (much faster)
        #     pos = 13  # Start after header
        #     while pos < total_size and pos < len(data):
        #         if pos + 1 >= len(data):
        #             break
        #
        #         opcode = data[pos]
        #         qualifier = data[pos + 1]
        #         pos += 2
        #
        #         if opcode == 0x04 and qualifier == 0x03:  # CONSTI  # noqa: PLR2004
        #             if pos + 3 >= len(data):
        #                 break
        #             value_offset = pos
        #             # Use struct.unpack for int32 (much faster than read_int32)
        #             const_value = struct.unpack(">i", data[pos:pos + 4])[0]
        #             pos += 4
        #             # Only cache positive values that could be StrRefs (negative are typically indices)
        #             if const_value >= 0:
        #                 location = f"offset_{value_offset}"
        #                 self._add_reference(const_value, identifier, location)
        #         elif opcode == 0x04:  # CONSTx  # noqa: PLR2004
        #             if qualifier == 0x04:  # CONSTF  # noqa: PLR2004
        #                 pos += 4
        #             elif qualifier == 0x05:  # CONSTS  # noqa: PLR2004
        #                 if pos + 1 >= len(data):
        #                     break
        #                 str_len = struct.unpack(">H", data[pos:pos + 2])[0]
        #                 pos += 2 + str_len
        #             elif qualifier == 0x06:  # CONSTO  # noqa: PLR2004
        #                 pos += 4
        #         elif opcode in (0x01, 0x03, 0x26, 0x27):  # CPDOWNSP, CPTOPSP, CPDOWNBP, CPTOPBP
        #             pos += 6
        #         elif opcode == 0x2C:  # STORE_STATE  # noqa: PLR2004
        #             pos += 8
        #         elif opcode in (0x1B, 0x1D, 0x1E, 0x1F, 0x23, 0x24, 0x25, 0x28, 0x29):
        #             pos += 4
        #         elif opcode == 0x05:  # ACTION  # noqa: PLR2004
        #             pos += 3
        #         elif opcode == 0x21:  # DESTRUCT  # noqa: PLR2004
        #             pos += 6
        #         elif opcode == 0x0B and qualifier == 0x24:  # EQUALTT  # noqa: PLR2004
        #             pos += 2
        #         elif opcode == 0x0C and qualifier == 0x24:  # NEQUALTT  # noqa: PLR2004
        #             pos += 2
        # except Exception:  # noqa: BLE001, S110
        #     pass

    def _scan_gff(
        self,
        identifier: ResourceIdentifier,
        gff_struct: GFFStruct,
        current_path: str = "",
    ) -> None:
        """Recursively scan GFF structure for LocalizedString fields with StrRefs.

        Args:
            identifier: Resource identifier being scanned
            gff_struct: GFF struct to scan
            current_path: Current field path as string (e.g., "FirstName" or "ItemList[0]")
        """
        for label, field_type, value in gff_struct:
            # Build field path using string concatenation (much faster than path operations)
            field_path: str = f"{current_path}.{label}" if current_path else label

            # LocalizedString fields
            is_locstring: bool = field_type == GFFFieldType.LocalizedString
            if is_locstring:
                locstring: LocalizedString = value  # type: ignore[assignment]
                if locstring.stringref != -1:
                    filename: str = f"{identifier.resname}.{identifier.restype.extension}"
                    _log_debug(f"Found StrRef {locstring.stringref} in GFF file '{filename}' at field path '{field_path}'")
                    self._add_reference(locstring.stringref, identifier, field_path)

            # Nested structs
            is_struct: bool = field_type == GFFFieldType.Struct
            if is_struct and isinstance(value, GFFStruct):
                self._scan_gff(identifier, value, field_path)

            # Lists
            is_list: bool = field_type == GFFFieldType.List
            if is_list and isinstance(value, GFFList):
                for idx, item in enumerate(value):
                    if isinstance(item, GFFStruct):
                        # Use string formatting for list indices (faster than path ops)
                        list_path: str = f"{field_path}[{idx}]"
                        self._scan_gff(identifier, item, list_path)

    def _add_reference(
        self,
        strref: int,
        identifier: ResourceIdentifier,
        location: str,
    ) -> None:
        """Add a StrRef reference to the cache.

        Uses O(1) dictionary lookup instead of O(n) linear search for performance.
        """
        filename: str = f"{identifier.resname}.{identifier.restype.extension}"

        # Track statistics
        self._total_references_found += 1
        self._files_with_strrefs.add(filename)

        # Initialize dict for this StrRef if needed
        if strref not in self._cache:
            self._cache[strref] = {}
            _log_verbose(f"  → Cached new StrRef {strref} from '{filename}' at '{location}'")

        # O(1) dictionary lookup instead of O(n) linear search
        if identifier in self._cache[strref]:
            # Identifier already exists, append location
            self._cache[strref][identifier].append(location)
        else:
            # New identifier for this StrRef
            self._cache[strref][identifier] = [location]

    def get_references(self, strref: int) -> list[tuple[ResourceIdentifier, list[str]]]:
        """Get all resources that reference a specific StrRef.

        Args:
            strref: StrRef to look up

        Returns:
            List of (resource_identifier, [field_paths]) tuples
        """
        # Convert dict format back to list of tuples for compatibility
        strref_dict = self._cache.get(strref, {})
        return [(identifier, locations) for identifier, locations in strref_dict.items()]

    def has_references(self, strref: int) -> bool:
        """Check if any resources reference this StrRef."""
        return strref in self._cache

    def get_statistics(self) -> dict[str, int]:
        """Get cache statistics.

        Returns:
            Dictionary with statistics about cached StrRef references
        """
        return {
            "unique_strrefs": len(self._cache),
            "total_references": self._total_references_found,
            "files_with_strrefs": len(self._files_with_strrefs),
        }

    def log_summary(self) -> None:
        """Log a summary of the cache contents."""
        stats: dict[str, int] = self.get_statistics()
        _log_verbose(
            f"\nStrRef Cache Summary:\n"
            f"  • {stats['unique_strrefs']} unique StrRefs cached\n"
            f"  • {stats['total_references']} total StrRef references found\n"
            f"  • {stats['files_with_strrefs']} files contain StrRef references"
        )

    def to_dict(self) -> dict[str, list[dict[str, str | list[str]]]]:
        """Serialize cache to dictionary for saving.

        Returns:
            Dictionary mapping strref -> list of {resname, restype, locations}
        """
        serialized: dict[str, list[dict[str, str | list[str]]]] = {}

        for strref, references_dict in self._cache.items():
            serialized[str(strref)] = [
                {
                    "resname": identifier.resname,
                    "restype": identifier.restype.extension,
                    "locations": locations,
                }
                for identifier, locations in references_dict.items()
            ]

        return serialized

    @classmethod
    def from_dict(
        cls,
        game: Game,
        data: dict[str, list[dict[str, str | list[str]]]],
    ) -> StrRefReferenceCache:
        """Restore cache from dictionary.

        Args:
            game: Game instance for cache initialization
            data: Serialized cache data

        Returns:
            Restored StrRefReferenceCache instance
        """
        from pykotor.extract.file import ResourceIdentifier  # noqa: PLC0415
        from pykotor.resource.type import ResourceType  # noqa: PLC0415

        cache = cls(game)

        for strref_str, references in data.items():
            strref: int = int(strref_str)
            cache._cache[strref] = {}

            for ref_data in references:
                resname: str = str(ref_data["resname"])
                restype_ext: str = str(ref_data["restype"])
                locations: list[str] = list(ref_data["locations"])  # type: ignore[arg-type]

                # Recreate ResourceIdentifier
                restype: ResourceType = ResourceType.from_extension(restype_ext)
                identifier = ResourceIdentifier(resname, restype)

                # Use dict assignment
                cache._cache[strref][identifier] = locations

                # Update statistics
                cache._total_references_found += len(locations)
                filename: str = f"{resname}.{restype_ext}"
                cache._files_with_strrefs.add(filename)

        _log_verbose(f"Restored StrRef cache from saved data: {len(cache._cache)} StrRefs, {cache._total_references_found} references")

        return cache


class TwoDAMemoryReferenceCache:
    """Cache of 2DA memory token references found during resource scanning.

    Maps (2da_filename, row_index) -> {resource_identifier: [field_paths]}
    where that row is referenced.

    This enables automatic generation of linking patches when 2DA rows are modified,
    similar to how StrRef linking works.
    """

    def __init__(self, game: Game):
        """Initialize cache.

        Args:
            game: Game instance (for potential game-specific logic)
        """
        self.game: Game = game

        # Map: (2da_filename, row_index) -> {resource_identifier: [field_paths]}
        # Using dict for O(1) lookups instead of list for O(n) linear search
        self._cache: dict[tuple[str, int], dict[ResourceIdentifier, list[str]]] = {}

        # Statistics
        self._total_references_found: int = 0
        self._files_with_2da_refs: set[str] = set()

    def scan_resource(
        self,
        resource: FileResource,
        data: bytes,
    ) -> None:
        """Scan a resource for 2DA memory references and cache them.

        Args:
            resource: FileResource being scanned
            data: Resource data bytes
        """
        identifier: ResourceIdentifier = resource.identifier()
        restype: ResourceType = resource.restype()

        try:
            # Only scan GFF files for 2DA references (use is_gff() for consistency with StrRef cache)
            if restype.is_gff():
                try:
                    gff_obj = read_gff(data)
                    self._scan_gff(identifier, gff_obj.root)
                except Exception:  # noqa: BLE001, S110
                    # Failed to parse GFF, skip
                    pass
            # All other file types don't contain 2DA references, skip silently

        except Exception:  # noqa: BLE001, S110
            # Skip files that fail to scan
            pass

    def _scan_gff(
        self,
        identifier: ResourceIdentifier,
        gff_struct: GFFStruct,
        current_path: str = "",
    ) -> None:
        """Recursively scan GFF structure for 2DA references.

        Args:
            identifier: Resource identifier
            gff_struct: GFF struct to scan
            current_path: Current field path as string (e.g., "Appearance" or "ItemList[0]")
        """
        # Get the mapping lazily to avoid circular dependency
        gff_field_to_2da_mapping = _get_gff_field_to_2da_mapping()

        for label, field_type, value in gff_struct:
            # Build field path using string concatenation (much faster than path operations)
            field_path: str = f"{current_path}.{label}" if current_path else label

            # Check if this field references a 2DA
            if label in gff_field_to_2da_mapping:
                # This field references a 2DA file
                twoda_identifier: ResourceIdentifier = gff_field_to_2da_mapping[label]
                twoda_filename = f"{twoda_identifier.resname}.{twoda_identifier.restype.extension}"

                # Extract the numeric value (row index)
                row_index: int | None = None
                if field_type in (
                    GFFFieldType.Int8,
                    GFFFieldType.Int16,
                    GFFFieldType.Int32,
                    GFFFieldType.Int64,
                ):
                    if isinstance(value, int):
                        row_index = value
                elif field_type in (
                    GFFFieldType.UInt8,
                    GFFFieldType.UInt16,
                    GFFFieldType.UInt32,
                    GFFFieldType.UInt64,
                ) and isinstance(value, int):
                    row_index = value

                if row_index is not None and row_index >= 0:
                    self._add_reference(twoda_filename, row_index, identifier, field_path)

            # Recurse into nested structures
            if field_type == GFFFieldType.Struct and isinstance(value, GFFStruct):
                self._scan_gff(identifier, value, field_path)
            elif field_type == GFFFieldType.List and isinstance(value, GFFList):
                for idx, item in enumerate(value):
                    if isinstance(item, GFFStruct):
                        # Use string formatting for list indices (faster than path ops)
                        list_path: str = f"{field_path}[{idx}]"
                        self._scan_gff(identifier, item, list_path)

    def _add_reference(
        self,
        twoda_filename: str,
        row_index: int,
        identifier: ResourceIdentifier,
        location: str,
    ) -> None:
        """Add a reference to the cache.

        Uses O(1) dictionary lookup instead of O(n) linear search for performance.

        Args:
            twoda_filename: Name of the 2DA file (e.g., "soundset.2da")
            row_index: Row index in the 2DA
            identifier: Resource identifier
            location: Field path in the GFF structure
        """
        key = (twoda_filename.lower(), row_index)
        filename: str = f"{identifier.resname}.{identifier.restype.extension}"

        # Track statistics
        self._total_references_found += 1
        self._files_with_2da_refs.add(filename)

        # Initialize dict for this 2DA row if needed
        if key not in self._cache:
            self._cache[key] = {}

        # O(1) dictionary lookup instead of O(n) linear search
        if identifier in self._cache[key]:
            # Identifier already exists, append location
            self._cache[key][identifier].append(location)
        else:
            # New identifier for this 2DA row
            self._cache[key][identifier] = [location]

    def get_references(
        self,
        twoda_filename: str,
        row_index: int,
    ) -> list[tuple[ResourceIdentifier, list[str]]]:
        """Get all references to a specific 2DA row.

        Args:
            twoda_filename: Name of the 2DA file
            row_index: Row index in the 2DA

        Returns:
            List of (resource_identifier, field_paths) tuples
        """
        key = (twoda_filename.lower(), row_index)
        # Convert dict format back to list of tuples for compatibility
        twoda_dict = self._cache.get(key, {})
        return [(identifier, locations) for identifier, locations in twoda_dict.items()]

    def has_references(
        self,
        twoda_filename: str,
        row_index: int,
    ) -> bool:
        """Check if any resources reference this 2DA row.

        Args:
            twoda_filename: Name of the 2DA file
            row_index: Row index in the 2DA

        Returns:
            True if references exist
        """
        key = (twoda_filename.lower(), row_index)
        return key in self._cache

    def get_statistics(self) -> dict[str, int]:
        """Get cache statistics.

        Returns:
            Dictionary with cache statistics
        """
        unique_2da_refs = len(self._cache)
        return {
            "unique_2da_refs": unique_2da_refs,
            "total_references": self._total_references_found,
            "files_with_2da_refs": len(self._files_with_2da_refs),
        }

    def log_summary(self) -> None:
        """Log a summary of cache contents."""
        stats = self.get_statistics()
        _log_verbose(f"2DA Memory Reference Cache: {stats['unique_2da_refs']} unique 2DA rows referenced")
        _log_verbose(f"  Total references: {stats['total_references']}")
        _log_verbose(f"  Files with 2DA refs: {stats['files_with_2da_refs']}")

    def to_dict(self) -> dict[str, list[dict[str, str | int | list[str]]]]:
        """Serialize cache to dictionary for saving.

        Returns:
            Serialized cache data
        """
        result: dict[str, list[dict[str, str | int | list[str]]]] = {}

        for (twoda_filename, row_index), references_dict in self._cache.items():
            key = f"{twoda_filename}:{row_index}"
            result[key] = [
                {
                    "resname": identifier.resname,
                    "restype": identifier.restype.extension,
                    "locations": locations,
                }
                for identifier, locations in references_dict.items()
            ]

        return result

    @classmethod
    def from_dict(
        cls,
        game: Game,
        data: dict[str, list[dict[str, str | int | list[str]]]],
    ) -> TwoDAMemoryReferenceCache:
        """Restore cache from serialized dictionary.

        Args:
            game: Game instance
            data: Serialized cache data

        Returns:
            Restored TwoDAMemoryReferenceCache
        """
        from pykotor.extract.file import ResourceIdentifier  # noqa: PLC0415
        from pykotor.resource.type import ResourceType  # noqa: PLC0415

        cache = cls(game)

        for key_str, references in data.items():
            # Parse key: "soundset.2da:123" -> ("soundset.2da", 123)
            twoda_filename, row_index_str = key_str.rsplit(":", 1)
            row_index = int(row_index_str)

            cache_key = (twoda_filename.lower(), row_index)
            cache._cache[cache_key] = {}

            for ref_data in references:
                resname: str = str(ref_data["resname"])
                restype_ext: str = str(ref_data["restype"])
                locations: list[str] = list(ref_data["locations"])  # type: ignore[arg-type]

                # Recreate ResourceIdentifier
                restype: ResourceType = ResourceType.from_extension(restype_ext)
                if restype is None or not restype.is_valid():
                    continue

                identifier = ResourceIdentifier(resname, restype)

                # Use dict assignment
                cache._cache[cache_key][identifier] = locations

                # Update statistics
                cache._total_references_found += len(locations)
                filename: str = f"{resname}.{restype_ext}"
                cache._files_with_2da_refs.add(filename)

        return cache


def find_all_strref_references(
    installation: Installation,
    strrefs: list[int],
    cache: StrRefReferenceCache | None = None,
    logger: Callable[[str], None] | None = None,
) -> tuple[dict[int, list[StrRefSearchResult]], StrRefReferenceCache]:
    """Find all references to multiple StrRefs in an installation using batch processing.

    This function scans resources once and finds references to all requested StrRefs,
    providing significant performance improvement over calling find_strref_references
    multiple times.

    Args:
        installation: The game installation to search
        strrefs: List of StrRef IDs to find references for
        cache: Optional pre-built StrRefReferenceCache (will build if not provided)
        logger: Optional logging function

    Returns:
    -------
        A tuple containing:
        - Dictionary mapping strref -> list of StrRefSearchResult
        - The StrRefReferenceCache instance used for the search
    """
    if not strrefs:
        return {}, cache or StrRefReferenceCache(installation.game())

    # Build cache if not provided
    if cache is None:
        if logger:
            logger(f"Building StrRef cache for {len(strrefs)} StrRefs for installation {installation.path()}...")
        cache = StrRefReferenceCache(installation.game())

        # Scan all resources to build the cache
        resource_count = 0
        skipped_count = 0
        last_logged_count = 0
        log_interval = 500  # Log progress every 500 resources

        for resource in installation:
            try:
                # Skip RIM files - they're not used at runtime
                try:
                    relative_path = resource.filepath().relative_to(installation.path())
                    path_parts = relative_path.parts
                    if path_parts and path_parts[0].lower() == "rims":
                        skipped_count += 1
                        continue
                except ValueError:
                    abs_path = resource.filepath()
                    if "rims" in abs_path.parts:
                        skipped_count += 1
                        continue

                # Filter by resource type: Only scan types that can contain StrRefs
                # This dramatically improves performance by skipping textures, models, audio, etc.
                restype = resource.restype()
                # StrRefs can only exist in: GFF files, 2DA files, SSF files, NCS files
                can_contain_strref = restype.is_gff() or restype == ResourceType.TwoDA or restype == ResourceType.SSF or restype == ResourceType.NCS
                if not can_contain_strref:
                    skipped_count += 1
                    continue

                data = resource.data()
                cache.scan_resource(resource, data)
                resource_count += 1

                # Log progress periodically
                if logger and resource_count - last_logged_count >= log_interval:
                    logger(f"  Scanning for StrRefs... {resource_count} resources processed for installation {installation.path()}")
                    last_logged_count = resource_count

            except Exception:  # noqa: BLE001, S110, S112
                skipped_count += 1
                continue

        if logger:
            logger(f"Cache built: scanned {resource_count} resources (skipped {skipped_count} files) for installation {installation.path()}")

    # Convert cache entries to StrRefSearchResult format
    results: dict[int, list[StrRefSearchResult]] = {}

    # Build a map of ResourceIdentifier -> FileResource by iterating installation ONCE
    # This is moved outside the strref loop to avoid re-iterating for each strref
    identifier_to_resource: dict[ResourceIdentifier, FileResource] = {}
    for res in installation:
        try:
            identifier_to_resource[res.identifier()] = res
        except Exception:  # noqa: BLE001, S110, S112
            continue

    for strref in strrefs:
        cache_entries = cache.get_references(strref)
        if not cache_entries:
            results[strref] = []
            continue

        # Convert cache format to StrRefSearchResult format
        strref_results: list[StrRefSearchResult] = []

        for identifier, location_strings in cache_entries:
            # Get the FileResource from our map
            try:
                found_resource: FileResource | None = identifier_to_resource.get(identifier)
                if found_resource:
                    # Convert location strings to proper location objects
                    locations: list[TwoDARefLocation | SSFRefLocation | GFFRefLocation | NCSRefLocation] = []

                    for loc_str in location_strings:
                        # Parse location string format: "row_12.name", "sound_Battlecry 1", "field_path", or byte offset
                        if loc_str.startswith("row_"):
                            # 2DA reference: "row_12.name"
                            parts = loc_str.replace("row_", "").split(".", 1)
                            if len(parts) == 2:
                                row_idx = int(parts[0])
                                column_name = parts[1]
                                locations.append(TwoDARefLocation(row_index=row_idx, column_name=column_name))
                        elif loc_str.startswith("sound_"):
                            # SSF reference: "sound_Battlecry 1"
                            sound_name = loc_str.replace("sound_", "")
                            try:
                                sound = SSFSound[sound_name]
                                locations.append(SSFRefLocation(sound=sound))
                            except KeyError:  # noqa: S110
                                pass
                        elif loc_str.startswith("offset_"):
                            # NCS reference: "offset_1234"
                            byte_offset = int(loc_str.replace("offset_", ""))
                            locations.append(NCSRefLocation(byte_offset=byte_offset))
                        else:
                            # GFF reference: field path
                            locations.append(GFFRefLocation(field_path=loc_str))

                    if locations:
                        strref_results.append(StrRefSearchResult(resource=found_resource, locations=locations))
            except Exception:  # noqa: BLE001, S110, S112
                continue

        results[strref] = strref_results

    return results, cache


def find_strref_references(
    installation: Installation,
    strref: int,
    cache: StrRefReferenceCache | None = None,
    logger: Callable[[str], None] | None = None,
) -> list[StrRefSearchResult]:
    """Find all references to a specific StrRef in an installation.

    This function scans 2DA, SSF, GFF, and NCS files for references to the given StrRef
    and returns complete location information for each reference.

    Args:
        installation: The game installation to search
        strref: The StrRef ID to find references for
        logger: Optional logging function

    Returns:
        List of StrRefSearchResult containing resources and their exact reference locations
    """
    K1Columns2DA, K2Columns2DA = _get_2da_columns()
    game = installation.game()

    def format_path_with_installation_prefix(path_str: str) -> str:
        """Format a relative path with the installation folder name as a prefix.

        Args:
            path_str: The relative path string (e.g., "data/2da.bif/planetary.2da")

        Returns:
            Formatted path with installation folder prefix (e.g., "swkotor/data/2da.bif/planetary.2da")
        """
        installation_path = installation.path()
        installation_folder = installation_path.name if installation_path else "unknown"
        return f"{installation_folder}/{path_str}"

    def resource_path_str(resource: FileResource, *, use_path_ident: bool = False) -> str:
        """Return resource path as installation-relative POSIX string when possible."""
        path_obj = resource.path_ident() if use_path_ident else resource.filepath()
        try:
            return path_obj.relative_to(installation.path()).as_posix()
        except ValueError:
            return str(path_obj)

    def formatted_resource_path(resource: FileResource, *, use_path_ident: bool = False) -> str:
        """Return formatted resource path prefixed with installation folder name."""
        return format_path_with_installation_prefix(resource_path_str(resource, use_path_ident=use_path_ident))

    def is_rims_resource(resource: FileResource) -> bool:
        """Check whether resource resides in RIMs (not used at runtime)."""
        try:
            parts = resource.filepath().relative_to(installation.path()).parts
            return bool(parts) and parts[0].lower() == "rims"
        except ValueError:
            return "rims" in resource.filepath().parts

    # Get game-specific 2DA columns that contain StrRefs
    strref_2da_columns: dict[str, set[str]] = {}
    if game.is_k1():
        strref_2da_columns = K1Columns2DA.StrRefs.as_dict()
    elif game.is_k2():
        strref_2da_columns = K2Columns2DA.StrRefs.as_dict()

    results: list[StrRefSearchResult] = []

    def scan_2da(resource: FileResource) -> StrRefSearchResult | None:
        """Scan a 2DA file and return locations where the StrRef is found."""
        try:
            twoda = read_2da(resource.data())
            filename = resource.filename().lower()

            if filename not in strref_2da_columns:
                return None

            locations: list[TwoDARefLocation] = []
            columns = strref_2da_columns[filename]

            for row_idx in range(twoda.get_height()):
                for column_name in columns:
                    if column_name == ">>##HEADER##<<":
                        continue

                    cell = twoda.get_cell(row_idx, column_name)
                    if cell and cell.strip().isdigit() and int(cell.strip()) == strref:
                        locations.append(TwoDARefLocation(row_index=row_idx, column_name=column_name))

                        if logger:
                            formatted_path = formatted_resource_path(resource, use_path_ident=True)
                            logger(f"    Found at: row {row_idx}, column '{column_name}' at {formatted_path}")

            if locations:
                return StrRefSearchResult(resource=resource, locations=locations)

        except Exception:  # noqa: BLE001, S110
            pass

        return None

    def scan_ssf(resource: FileResource) -> StrRefSearchResult | None:
        """Scan an SSF file and return locations where the StrRef is found."""
        try:
            ssf = read_ssf(resource.data())
            locations: list[SSFRefLocation] = []

            for sound in SSFSound:
                sound_strref = ssf.get(sound)
                if sound_strref == strref:
                    locations.append(SSFRefLocation(sound=sound))

                    if logger:
                        formatted_path = formatted_resource_path(resource, use_path_ident=True)
                        logger(f"    Found at: sound slot '{sound.name}' at {formatted_path}")

            if locations:
                return StrRefSearchResult(resource=resource, locations=locations)

        except Exception:  # noqa: BLE001, S110
            pass

        return None

    def scan_gff(resource: FileResource) -> StrRefSearchResult | None:
        """Scan a GFF file and return locations where the StrRef is found."""
        try:
            gff = read_gff(resource.data())
            locations: list[GFFRefLocation] = []

            def recurse_gff(gff_struct: GFFStruct, path_prefix: str = "") -> None:
                """Recursively scan GFF struct."""
                for field_label, ftype, fval in gff_struct:
                    field_path = f"{path_prefix}.{field_label}" if path_prefix else field_label

                    try:
                        # Check LocalizedString fields
                        if ftype == GFFFieldType.LocalizedString and isinstance(fval, LocalizedString):
                            if fval.stringref == strref:
                                locations.append(GFFRefLocation(field_path=field_path))

                        # Recurse into nested structs
                        elif ftype == GFFFieldType.Struct and isinstance(fval, GFFStruct):
                            recurse_gff(fval, field_path)

                        # Recurse into list items
                        elif ftype == GFFFieldType.List and isinstance(fval, GFFList):
                            for idx, item in enumerate(fval):
                                if isinstance(item, GFFStruct):
                                    list_path = f"{field_path}[{idx}]"
                                    recurse_gff(item, list_path)
                    except Exception as e:  # noqa: BLE001
                        # Individual field errors - log but continue processing other fields
                        if logger:
                            path_str = resource_path_str(resource)
                            logger(f"[Debug] Error processing GFF field '{field_path}' in {path_str}: {e.__class__.__name__}: {e}")
                        print(traceback.format_exc())

            recurse_gff(gff.root)

            if locations and logger:
                formatted_path = formatted_resource_path(resource, use_path_ident=True)
                for location in locations:
                    logger(f"    Found at: field path '{location.field_path}' at {formatted_path}")

            if locations:
                return StrRefSearchResult(resource=resource, locations=locations)

        except Exception as e:  # noqa: BLE001, S110
            # Log which file failed to parse, but don't spam on expected failures (non-GFF files)
            if logger:
                path_str = resource_path_str(resource)
                logger(f"[Debug] Error scanning GFF {resource.filename()}: {e.__class__.__name__}: {e}")
            print(traceback.format_exc())

        return None

    def scan_ncs(resource: FileResource) -> StrRefSearchResult | None:
        """Scan an NCS file and return byte offsets where the StrRef is found in CONSTI instructions."""
        try:
            locations: list[NCSRefLocation] = []
            ncs_data = resource.data()

            # Parse NCS bytecode to find CONSTI instructions
            with BinaryReader.from_auto(ncs_data) as reader:
                if reader.read_string(4) != "NCS ":
                    return None
                if reader.read_string(4) != "V1.0":
                    return None
                magic_byte = reader.read_uint8()
                if magic_byte != 0x42:  # noqa: PLR2004
                    return None
                total_size = reader.read_uint32(big=True)

                # OPTIMIZATION: Removed redundant remaining() check - if position() < total_size,
                # then remaining() = total_size - position() > 0. This eliminates ~902k redundant
                # position() calls (one per loop iteration via remaining()).
                while reader.position() < total_size:
                    opcode = reader.read_uint8()
                    qualifier = reader.read_uint8()

                    if opcode == 0x04 and qualifier == 0x03:  # CONSTI  # noqa: PLR2004
                        value_offset = reader.position()
                        const_value = reader.read_int32(big=True)
                        if const_value == strref:
                            locations.append(NCSRefLocation(byte_offset=value_offset))
                    elif opcode == 0x04:  # CONSTx  # noqa: PLR2004
                        if qualifier == 0x04:  # CONSTF  # noqa: PLR2004
                            reader.skip(4)
                        elif qualifier == 0x05:  # CONSTS  # noqa: PLR2004
                            str_len = reader.read_uint16(big=True)
                            reader.skip(str_len)
                        elif qualifier == 0x06:  # CONSTO  # noqa: PLR2004
                            reader.skip(4)
                    elif opcode in (0x01, 0x03, 0x26, 0x27):  # CPDOWNSP, CPTOPSP, CPDOWNBP, CPTOPBP
                        reader.skip(6)
                    elif opcode == 0x2C:  # STORE_STATE  # noqa: PLR2004
                        reader.skip(8)
                    elif opcode in (0x1B, 0x1D, 0x1E, 0x1F, 0x23, 0x24, 0x25, 0x28, 0x29):
                        reader.skip(4)
                    elif opcode == 0x05:  # ACTION  # noqa: PLR2004
                        reader.skip(3)
                    elif opcode == 0x21:  # DESTRUCT  # noqa: PLR2004
                        reader.skip(6)
                    elif opcode == 0x0B and qualifier == 0x24:  # EQUALTT  # noqa: PLR2004
                        reader.skip(2)
                    elif opcode == 0x0C and qualifier == 0x24:  # NEQUALTT  # noqa: PLR2004
                        reader.skip(2)

            if locations:
                if logger:
                    formatted_path = formatted_resource_path(resource, use_path_ident=True)
                    for location in locations:
                        logger(f"    Found at: byte offset {location.byte_offset:#X} (0x{location.byte_offset:X}) at {formatted_path}")

                return StrRefSearchResult(resource=resource, locations=locations)

        except Exception:  # noqa: BLE001, S110
            pass

        return None

    # If cache is provided, use it for faster lookup
    if cache is not None:
        cache_entries = cache.get_references(strref)
        if cache_entries:
            # Build a map of ResourceIdentifier -> FileResource by iterating installation
            identifier_to_resource: dict[ResourceIdentifier, FileResource] = {}
            for res in installation:
                try:
                    identifier_to_resource[res.identifier()] = res
                except Exception:  # noqa: BLE001, S110, S112
                    continue

            # Convert cache format to StrRefSearchResult format
            for identifier, location_strings in cache_entries:
                try:
                    found_resource: FileResource | None = identifier_to_resource.get(identifier)
                    if found_resource is None:
                        continue

                    # Convert location strings to proper location objects
                    locations: list[TwoDARefLocation | SSFRefLocation | GFFRefLocation | NCSRefLocation] = []

                    for loc_str in location_strings:
                        if loc_str.startswith("row_"):
                            parts = loc_str.replace("row_", "").split(".", 1)
                            if len(parts) == 2:
                                row_idx = int(parts[0])
                                column_name = parts[1]
                                locations.append(TwoDARefLocation(row_index=row_idx, column_name=column_name))
                        elif loc_str.startswith("sound_"):
                            sound_name = loc_str.replace("sound_", "")
                            try:
                                sound = SSFSound[sound_name]
                                locations.append(SSFRefLocation(sound=sound))
                            except KeyError:  # noqa: S110
                                pass
                        elif loc_str.startswith("offset_"):
                            byte_offset = int(loc_str.replace("offset_", ""))
                            locations.append(NCSRefLocation(byte_offset=byte_offset))
                        else:
                            locations.append(GFFRefLocation(field_path=loc_str))

                    if locations:
                        results.append(StrRefSearchResult(resource=found_resource, locations=locations))
                except Exception:  # noqa: BLE001, S110, S112
                    continue

            return results

    # Scan all resources in the installation (no cache available)
    for resource in installation:
        restype = resource.restype()

        # Skip RIM files - they're not used at runtime
        if is_rims_resource(resource):
            continue

        # Check 2DA files
        if restype == ResourceType.TwoDA:
            result = scan_2da(resource)
            if result:
                results.append(result)

        # Check SSF files
        elif restype == ResourceType.SSF:
            result = scan_ssf(resource)
            if result:
                results.append(result)

        # Check NCS files - FIXME: TEMPORARILY DISABLED (will revisit later)
        # elif restype == ResourceType.NCS:
        #     result = scan_ncs(resource)
        #     if result:
        #         results.append(result)

        # Try to check as GFF - ONLY if it has a GFF extension
        elif restype.extension in GFFContent.get_extensions():
            result = scan_gff(resource)
            if result:
                results.append(result)
        # Skip all other resource types (textures, audio, etc.)

    return results


def find_tlk_entry_references(
    installation: Installation,
    query_stringref: int,
    order: list[SearchLocation] | None = None,
    *,
    capsules: list[Capsule] | None = None,
    folders: list[Path] | None = None,
    logger: Callable[[str], None] | None = None,
) -> set[FileResource]:
    """Finds all gffs that utilize this stringref in their localizedstring.

    If no gffs could not be found the value will return None.

    Args:
    ----
        installation: The installation to search
        query_stringref: A number representing the locstring to find.
        order: The ordered list of locations to check.
        capsules: An extra list of capsules to search in.
        folders: An extra list of folders to search in.
        logger: A logger to use for logging. (Optional)

    Returns:
    -------
        A set of FileResources.
    """
    capsules = [] if capsules is None else capsules
    folders = [] if folders is None else folders
    if order is None:
        order = [
            SearchLocation.CUSTOM_FOLDERS,
            SearchLocation.OVERRIDE,
            SearchLocation.CUSTOM_MODULES,
            SearchLocation.CHITIN,
            SearchLocation.MODULES,
        ]

    found_resources: set[FileResource] = set()
    gff_extensions: set[str] = GFFContent.get_extensions()
    relevant_2da_filenames: dict[str, set[str]] = {}

    if installation.game().is_k1():  # TODO(th3w1zard1): TSL:
        relevant_2da_filenames = K1Columns2DA.StrRefs.as_dict()
    elif installation.game().is_k2():
        relevant_2da_filenames = K2Columns2DA.StrRefs.as_dict()

    installation_path = installation.path()

    def _resource_path_str(resource: FileResource) -> str:
        """Return resource path as installation-relative POSIX string when possible."""
        try:
            return resource.filepath().relative_to(installation_path).as_posix()
        except ValueError:
            return str(resource.filepath())

    def check_2da(resource2da: FileResource) -> bool:
        valid_2da: TwoDA | None = None
        with suppress(ValueError, OSError):
            valid_2da = read_2da(resource2da.data())
        if not valid_2da:
            print(f"'{resource2da._path_ident_obj}' cannot be loaded, probably corrupted.")  # noqa: SLF001
            return False
        filename_2da = resource2da.filename().lower()

        path_str = _resource_path_str(resource2da)

        found_locations: list[tuple[str, int | None]] = []  # (column_name, row_index)

        for column_name in relevant_2da_filenames[filename_2da]:
            if column_name == ">>##HEADER##<<":
                for header in valid_2da.get_headers():
                    try:
                        stripped_header = header.strip()
                        if not stripped_header.isdigit():
                            if stripped_header and stripped_header not in ("****", "*****", "-1"):
                                RobustLogger().warning(f"header '{header}' in '{filename_2da}' is invalid, expected a stringref number.")
                            continue
                        if int(stripped_header) == query_stringref:
                            found_locations.append((">>##HEADER##<<", None))
                            if logger:
                                logger(f"    Found at: header '{header}' at {path_str}")
                    except Exception as e:  # noqa: BLE001
                        RobustLogger().error("Error parsing '%s' header '%s': %s", filename_2da, header, str(e), exc_info=False)
            else:
                try:
                    for i, cell in enumerate(valid_2da.get_column(column_name)):
                        stripped_cell = cell.strip()
                        if not stripped_cell.isdigit():
                            if stripped_cell and stripped_cell not in ("****", "*****", "-1"):
                                RobustLogger().warning(
                                    f"column '{column_name}' rowindex {i} in '{filename_2da}' is invalid, expected a stringref number. Instead got '{cell}'"
                                )
                            continue
                        if int(stripped_cell) == query_stringref:
                            found_locations.append((column_name, i))
                            if logger:
                                logger(f"    Found at: row {i}, column '{column_name}' at {path_str}")
                except Exception as e:  # noqa: BLE001
                    RobustLogger().error("Error parsing '%s' column '%s': %s", filename_2da, column_name, str(e), exc_info=False)

        return len(found_locations) > 0

    def check_ssf(resource_ssf: FileResource) -> bool:
        """Check if an SSF file contains the query StrRef."""
        valid_ssf: SSF | None = None
        with suppress(ValueError, OSError):
            valid_ssf = read_ssf(resource_ssf.data())
        if not valid_ssf:
            return False

        path_str = _resource_path_str(resource_ssf)

        # Check all sound slots for this StrRef
        for sound in SSFSound:
            sound_strref = valid_ssf.get(sound)
            if sound_strref == query_stringref:
                if logger:
                    logger(f"    Found at: sound slot '{sound.name}' at {path_str}")
                return True
        return False

    def check_ncs(resource_ncs: FileResource) -> bool:
        """Check if an NCS (compiled script) file contains the query StrRef.

        StrRefs in scripts appear as CONSTI (constant integer) instructions.
        We extract all CONSTI values and check if any match the query.
        """
        valid_ncs: NCS | None = None
        with suppress(ValueError, OSError):
            valid_ncs = read_ncs(resource_ncs.data())
        if not valid_ncs:
            return False

        path_str = _resource_path_str(resource_ncs)

        # Get byte offsets using the existing function
        offsets = get_ncs_consti_offsets(resource_ncs, query_stringref)

        # Log found offsets
        if offsets and logger:
            for offset in offsets:
                logger(f"    Found at: byte offset {offset:#X} (0x{offset:X}) at {path_str}")

        return len(offsets) > 0

    def check_gff(resource_gff: FileResource) -> bool:
        """Check if a GFF file contains the query StrRef."""
        valid_gff: GFF | None = None
        with suppress(ValueError, OSError):
            valid_gff = read_gff(resource_gff.data())
        if valid_gff is None:
            return False

        path_str = _resource_path_str(resource_gff)

        # Track found field paths for logging
        found_paths: list[str] = []

        def recurse_gff_structs_with_logging(gff_struct: GFFStruct, path_prefix: str = "") -> bool:
            """Recursively scan GFF struct and collect field paths."""
            for field_label, ftype, fval in gff_struct:
                field_path = f"{path_prefix}.{field_label}" if path_prefix else field_label

                if ftype == GFFFieldType.List and isinstance(fval, GFFList):
                    for idx, list_struct in enumerate(fval):
                        list_path = f"{field_path}[{idx}]"
                        if recurse_gff_structs_with_logging(list_struct, list_path):
                            found_paths.append(list_path)
                if ftype == GFFFieldType.Struct and isinstance(fval, GFFStruct) and recurse_gff_structs_with_logging(fval, field_path):
                    found_paths.append(field_path)
                if ftype != GFFFieldType.LocalizedString or not isinstance(fval, LocalizedString):
                    continue
                if fval.stringref == query_stringref:
                    found_paths.append(field_path)
                    return True
            return False

        result = recurse_gff_structs_with_logging(valid_gff.root)

        # Log all found field paths
        if result and logger:
            for field_path in found_paths:
                logger(f"    Found at: field path '{field_path}' at {path_str}")

        return result

    def get_ncs_consti_offsets(resource_ncs: FileResource, target_value: int) -> list[int]:
        """Get byte offsets of all CONSTI instructions with a specific value.

        Returns list of byte offsets where the 4-byte integer value starts
        (i.e., offset + 2 from instruction start).
        """
        offsets: list[int] = []
        try:
            # Re-read with offset tracking
            with BinaryReader.from_auto(resource_ncs.data()) as reader:
                # Skip NCS header (13 bytes)
                if reader.read_string(4) != "NCS ":
                    return offsets
                if reader.read_string(4) != "V1.0":
                    return offsets
                magic_byte = reader.read_uint8()
                if magic_byte != 0x42:  # noqa: PLR2004
                    return offsets
                total_size = reader.read_uint32(big=True)

                # Now read instructions and track offsets
                while reader.position() < total_size and reader.remaining() > 0:
                    opcode = reader.read_uint8()
                    qualifier = reader.read_uint8()

                    # Check if this is CONSTI (opcode=0x04, qualifier=0x03)
                    if opcode == 0x04 and qualifier == 0x03:  # CONSTI  # noqa: PLR2004
                        value_offset = reader.position()  # Current position is where the 4-byte value starts
                        const_value = reader.read_int32(big=True)
                        if const_value == target_value:
                            offsets.append(value_offset)
                    # Skip to next instruction based on opcode/qualifier
                    # This is simplified - just skip common patterns
                    elif opcode == 0x04:  # CONSTx  # noqa: PLR2004
                        if qualifier == 0x04:  # CONSTF  # noqa: PLR2004
                            reader.skip(4)
                        elif qualifier == 0x05:  # CONSTS  # noqa: PLR2004
                            str_len = reader.read_uint16(big=True)
                            reader.skip(str_len)
                        elif qualifier == 0x06:  # CONSTO  # noqa: PLR2004
                            reader.skip(4)
                    elif opcode in (0x01, 0x03, 0x26, 0x27):  # CPDOWNSP, CPTOPSP, CPDOWNBP, CPTOPBP
                        reader.skip(6)
                    elif opcode == 0x2C:  # STORE_STATE  # noqa: PLR2004
                        reader.skip(8)
                    elif opcode in (0x1B, 0x1D, 0x1E, 0x1F, 0x23, 0x24, 0x25, 0x28, 0x29):  # MOVSP, jumps, inc/dec
                        reader.skip(4)
                    elif opcode == 0x05:  # ACTION  # noqa: PLR2004
                        reader.skip(3)
                    elif opcode == 0x21:  # DESTRUCT  # noqa: PLR2004
                        reader.skip(6)
                    elif opcode == 0x0B and qualifier == 0x24:  # EQUALTT  # noqa: PLR2004
                        reader.skip(2)
                    elif opcode == 0x0C and qualifier == 0x24:  # NEQUALTT  # noqa: PLR2004
                        reader.skip(2)
                        # Other instructions have no additional data

        except Exception:  # noqa: BLE001, S110
            # If anything fails, return what we found so far
            pass

        return offsets

    def recurse_gff_lists(gff_list: GFFList) -> bool:
        for gff_struct in gff_list:
            result = recurse_gff_structs(gff_struct)
            if result:
                return True
        return False

    def recurse_gff_structs(gff_struct: GFFStruct) -> bool:
        for _label, ftype, fval in gff_struct:
            if ftype == GFFFieldType.List and isinstance(fval, GFFList):
                result = recurse_gff_lists(fval)
                if result:
                    return True
            if ftype == GFFFieldType.Struct and isinstance(fval, GFFStruct):
                result = recurse_gff_structs(fval)
                if result:
                    return True
            if ftype != GFFFieldType.LocalizedString or not isinstance(fval, LocalizedString):
                continue
            if fval.stringref == query_stringref:  # the matching strref was found
                return True
        return False

    def try_get_gff(gff_data: bytes) -> GFF | None:
        with suppress(OSError, ValueError):
            return read_gff(gff_data)
        return None

    def check_dict(resource_dict: dict[str, list[FileResource]]):
        for resources in resource_dict.values():
            check_list(resources)

    def check_list(resource_list: list[FileResource]):
        for resource in resource_list:
            this_restype: ResourceType = resource.restype()

            # Check 2DA files
            if resource.filename().lower() in relevant_2da_filenames and this_restype == ResourceType.TwoDA and check_2da(resource):
                found_resources.add(resource)
                continue

            # Check SSF files
            if this_restype == ResourceType.SSF and check_ssf(resource):
                found_resources.add(resource)
                continue

            # Check NCS files - FIXME: TEMPORARILY DISABLED (will revisit later)
            # if this_restype == ResourceType.NCS and check_ncs(resource):
            #     found_resources.add(resource)
            #     continue

            # Check GFF files
            if this_restype.extension not in gff_extensions:
                continue
            valid_gff: GFF | None = try_get_gff(resource.data())
            if valid_gff is None:
                continue
            if not recurse_gff_structs(valid_gff.root):
                continue
            found_resources.add(resource)

    def check_capsules(capsules_list: list[Capsule]):
        for capsule in capsules_list:
            for resource in capsule.resources():
                this_restype: ResourceType = resource.restype()

                # Check 2DA files
                if resource.filename().lower() in relevant_2da_filenames and this_restype == ResourceType.TwoDA and check_2da(resource):
                    found_resources.add(resource)
                    continue

                # Check SSF files
                if this_restype == ResourceType.SSF and check_ssf(resource):
                    found_resources.add(resource)
                    continue

                # Check NCS files - FIXME: TEMPORARILY DISABLED (will revisit later)
                # if this_restype == ResourceType.NCS and check_ncs(resource):
                #     found_resources.add(resource)
                #     continue

                # Check GFF files
                if this_restype.extension not in gff_extensions:
                    continue
                valid_gff: GFF | None = try_get_gff(resource.data())
                if valid_gff is None:
                    continue
                if not recurse_gff_structs(valid_gff.root):
                    continue
                found_resources.add(resource)

    def check_folders(values: list[Path]):
        relevant_files: set[Path] = set()
        for folder in values:  # Having two loops makes it easier to filter out irrelevant files when stepping through the 2nd
            relevant_files.update(
                file
                for file in folder.rglob("*")
                if (
                    file.suffix
                    and (
                        file.suffix[1:].casefold() in gff_extensions
                        or (file.name.lower() in relevant_2da_filenames and file.suffix.casefold() == ".2da")
                        or file.suffix.casefold() == ".ssf"
                        or file.suffix.casefold() == ".ncs"
                    )
                    and file.is_file()
                )
            )
        for gff_file in relevant_files:
            restype: ResourceType | None = ResourceType.from_extension(gff_file.suffix)
            if not restype:
                continue
            fileres = FileResource(resname=gff_file.stem, restype=restype, size=gff_file.stat().st_size, offset=0, filepath=gff_file)

            # Check 2DA files
            if restype == ResourceType.TwoDA and check_2da(fileres):
                found_resources.add(fileres)
                continue

            # Check SSF files
            if restype == ResourceType.SSF and check_ssf(fileres):
                found_resources.add(fileres)
                continue

            # Check NCS files - FIXME: TEMPORARILY DISABLED (will revisit later)
            # if restype == ResourceType.NCS and check_ncs(fileres):
            #     found_resources.add(fileres)
            #     continue

            # Check GFF files
            gff_data = BinaryReader.load_file(gff_file)
            valid_gff: GFF | None = None
            with suppress(ValueError, OSError):
                valid_gff = read_gff(gff_data)
            if not valid_gff:
                continue
            if not recurse_gff_structs(valid_gff.root):
                continue
            found_resources.add(fileres)

    # Access Installation internal properties (they're properties so they'll work)
    function_map: dict[SearchLocation, Callable] = {
        SearchLocation.OVERRIDE: lambda: check_dict(installation._override),  # noqa: SLF001
        SearchLocation.MODULES: lambda: check_dict(installation._modules),  # noqa: SLF001
        SearchLocation.RIMS: lambda: check_dict(installation._rims),  # noqa: SLF001
        SearchLocation.CHITIN: lambda: check_list(installation._chitin) or check_list(installation._patch_erf),  # noqa: SLF001
        SearchLocation.CUSTOM_MODULES: lambda: check_capsules(capsules),
        SearchLocation.CUSTOM_FOLDERS: lambda: check_folders(folders),  # type: ignore[arg-type]
    }

    for item in order:
        assert isinstance(item, SearchLocation), f"{type(item).__name__}: {item}"
        function_map.get(item, lambda: None)()

    return found_resources



"""Pack command implementation.

This module handles packing source files into KOTOR module/ERF/hak files.
"""

from __future__ import annotations

import glob

from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from argparse import Namespace
    from logging import Logger

from pykotor.cli.cfg_parser import load_config
from pykotor.cli.commands.compile import cmd_compile
from pykotor.cli.commands.convert import cmd_convert
from pykotor.resource.formats.erf import ERF, ERFType, write_erf
from pykotor.resource.type import ResourceType


def should_overwrite_file(file_path: Path, source_newest: float, mode: str, logger: Logger) -> bool:
    """Determine if file should be overwritten based on mode.

    Args:
    ----
        file_path: Path to the existing file
        source_newest: Timestamp of newest source file
        mode: Overwrite mode (ask/default/always/never)
        logger: Logger instance

    Returns:
    -------
        True if file should be overwritten
    """
    if mode == "always":
        return True
    if mode == "never":
        return False

    existing_time = file_path.stat().st_mtime
    if mode == "default":
        # Overwrite if existing is older than source
        return existing_time < source_newest

    # mode == "ask"
    logger.info(f"File exists: {file_path}")
    logger.info(f"  Existing: {existing_time}")
    logger.info(f"  Source:   {source_newest}")
    response = input("Overwrite? (yes/no): ").strip().lower()
    return response in ("yes", "y")


def cmd_pack(args: Namespace, logger: Logger) -> int:
    """Handle pack command - pack sources into module/erf/hak.

    Args:
    ----
        args: Parsed command line arguments
        logger: Logger instance

    Returns:
    -------
        Exit code (0 for success, non-zero for error)
    """
    # Load configuration
    config = load_config(logger)
    if config is None:
        return 1

    # Determine targets
    target_names = args.targets if args.targets else [None]
    if "all" in target_names:
        targets = config.targets
    else:
        targets = []
        for name in target_names:
            target = config.get_target(name)
            if target is None:
                if name:
                    logger.error(f"Target not found: {name}")
                else:
                    logger.error("No default target found")
                return 1
            targets.append(target)

    # Process each target
    for target in targets:
        target_name = target.get("name", "unnamed")
        logger.info(f"Packing target: {target_name}")

        # Run convert unless --noConvert
        if not getattr(args, "noConvert", False):
            logger.info("Converting JSON to GFF...")
            convert_result = cmd_convert(args, logger)
            if convert_result != 0:
                logger.error("Convert failed")
                return convert_result

        # Run compile unless --noCompile
        if not getattr(args, "noCompile", False):
            logger.info("Compiling scripts...")
            compile_result = cmd_compile(args, logger)
            if compile_result != 0 and getattr(args, "abortOnCompileError", False):
                logger.error("Compile failed, aborting pack")
                return compile_result

        # Get cache directory
        cache_dir = config.root_dir / ".pykotorcli" / "cache" / target_name
        if not cache_dir.exists():
            logger.error(f"Cache directory not found: {cache_dir}")
            logger.info("Run 'pykotorcli convert' and 'pykotorcli compile' first")
            return 1

        # Determine output file
        if hasattr(args, "pack_file") and args.pack_file:
            output_file = Path(args.pack_file)
        else:
            output_filename = config.resolve_target_value(target, "file")
            if not output_filename:
                logger.error("No output file specified for target")
                return 1
            output_file = config.root_dir / output_filename

        logger.info(f"Output file: {output_file}")

        # Check if output file exists
        if output_file.exists():
            # Find newest source file
            sources = config.get_target_sources(target)
            include_patterns = sources.get("include", [])
            newest_source_time = 0.0
            for pattern in include_patterns:
                pattern_path = str(config.root_dir / pattern)
                matches = glob.glob(pattern_path, recursive=True)
                for match in matches:
                    match_path = Path(match)
                    if match_path.is_file():
                        mtime = match_path.stat().st_mtime
                        newest_source_time = max(newest_source_time, mtime)

            overwrite_mode = getattr(args, "overwritePackedFile", "ask")
            if not should_overwrite_file(output_file, newest_source_time, overwrite_mode, logger):
                logger.info("Keeping existing file, skipping pack")
                continue

        # Collect files from cache
        cache_files = list(cache_dir.glob("**/*"))
        cache_files = [f for f in cache_files if f.is_file()]

        # Apply filter patterns
        sources = config.get_target_sources(target)
        filter_patterns = sources.get("filter", [])
        filtered_files = []
        for cache_file in cache_files:
            import fnmatch

            should_filter = False
            for pattern in filter_patterns:
                if fnmatch.fnmatch(cache_file.name, pattern):
                    should_filter = True
                    logger.debug(f"Filtering out: {cache_file.name}")
                    break
            if not should_filter:
                filtered_files.append(cache_file)

        logger.info(f"Packing {len(filtered_files)} files")

        # Determine archive type
        suffix = output_file.suffix.lower()
        if suffix in (".mod", ".erf", ".sav"):
            # Create ERF archive
            try:
                erf = ERF()
                if suffix == ".mod":
                    erf.erf_type = ERFType.MOD
                elif suffix == ".sav":
                    erf.erf_type = ERFType.SAV
                else:
                    erf.erf_type = ERFType.ERF

                # Add files to ERF
                for cache_file in filtered_files:
                    # Determine resref and restype
                    name_parts = cache_file.stem.split(".")
                    if len(name_parts) >= 2:
                        resref = ".".join(name_parts[:-1])
                        ext = name_parts[-1]
                    else:
                        resref = cache_file.stem
                        ext = cache_file.suffix.lstrip(".")

                    # Get ResourceType
                    try:
                        restype = ResourceType.from_extension(ext)
                    except ValueError:
                        logger.warning(f"Unknown resource type for {cache_file.name}, skipping")
                        continue

                    # Read file data
                    file_data = cache_file.read_bytes()

                    # Add to ERF
                    erf.set(resref, restype, file_data)

                # Write ERF
                output_file.parent.mkdir(parents=True, exist_ok=True)
                write_erf(erf, output_file)
                logger.info(f"Successfully packed {output_file}")

            except Exception as e:
                logger.exception(f"Failed to pack ERF: {e}")
                return 1

        elif suffix == ".rim":
            logger.error("RIM packing not yet implemented")
            return 1
        else:
            logger.error(f"Unsupported output file type: {suffix}")
            return 1

    return 0



"""Enhanced type system for NCS format based on DeNCS implementation.

This module provides a comprehensive type system for NCS bytecode operations,
including basic types, compound types, and struct types with proper size calculations.
"""

from __future__ import annotations
from pykotor.resource.formats._base import BiowareResource

from enum import IntEnum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Self


class NCSTypeCode(IntEnum):
    """Type codes used in NCS bytecode operations."""

    # Basic types
    NONE = 0x00
    STACK = 0x01
    INTEGER = 0x03
    FLOAT = 0x04
    STRING = 0x05
    OBJECT = 0x06

    # Engine types
    EFFECT = 0x10
    EVENT = 0x11
    LOCATION = 0x12
    TALENT = 0x13

    # Compound types
    INTINT = 0x20
    FLOATFLOAT = 0x21
    OBJECTOBJECT = 0x22
    STRINGSTRING = 0x23
    STRUCTSTRUCT = 0x24
    INTFLOAT = 0x25
    FLOATINT = 0x26

    # Compound engine types
    EFFECTEFFECT = 0x30
    EVENTEVENT = 0x31
    LOCLOC = 0x32
    TALTAL = 0x33

    # Vector types
    VECTORVECTOR = 0x3A
    VECTORFLOAT = 0x3B
    FLOATVECTOR = 0x3C

    # Special types
    VECTOR = 0xF0  # -16 in signed byte
    STRUCT = 0xF1  # -15 in signed byte
    INVALID = 0xFF  # -1 in signed byte


class NCSType(BiowareResource):
    """Represents a type in the NCS type system.

    Provides type information, size calculations, and type conversion utilities.
    Based on DeNCS Type.java implementation.
    """

    __slots__ = ("_size", "_type_code")

    def __init__(self, type_code: int | NCSTypeCode | str):
        """Initialize an NCS type.

        Args:
        ----
            type_code: Type code as integer, NCSTypeCode enum, or string name
        """
        if isinstance(type_code, str):
            self._type_code = self._decode_string(type_code)
        elif isinstance(type_code, NCSTypeCode):
            self._type_code = type_code
        else:
            self._type_code = NCSTypeCode(type_code)

        self._size = self._calculate_size()

    @staticmethod
    def _decode_string(type_str: str) -> NCSTypeCode:
        """Decode a type string to a type code.

        Args:
        ----
            type_str: String representation of type (e.g., "int", "float")

        Returns:
        -------
            NCSTypeCode: Corresponding type code

        Raises:
        ------
            ValueError: If type string is unknown
        """
        type_map = {
            "void": NCSTypeCode.NONE,
            "int": NCSTypeCode.INTEGER,
            "float": NCSTypeCode.FLOAT,
            "string": NCSTypeCode.STRING,
            "object": NCSTypeCode.OBJECT,
            "effect": NCSTypeCode.EFFECT,
            "event": NCSTypeCode.EVENT,
            "location": NCSTypeCode.LOCATION,
            "talent": NCSTypeCode.TALENT,
            "vector": NCSTypeCode.VECTOR,
            "action": NCSTypeCode.NONE,  # Actions have void type
            "struct": NCSTypeCode.STRUCT,
            "stack": NCSTypeCode.STACK,
            # Aliases
            "INT": NCSTypeCode.INTEGER,
            "OBJECT_ID": NCSTypeCode.OBJECT,
        }

        if type_str not in type_map:
            msg = f"Unknown type string: {type_str}"
            raise ValueError(msg)

        return type_map[type_str]

    def _calculate_size(self) -> int:
        """Calculate the size of this type in 4-byte words.

        Returns:
        -------
            int: Size in 4-byte words

        Raises:
        ------
            ValueError: If type size cannot be determined
        """
        size_map = {
            NCSTypeCode.NONE: 0,
            NCSTypeCode.STACK: 1,
            NCSTypeCode.INTEGER: 1,
            NCSTypeCode.FLOAT: 1,
            NCSTypeCode.STRING: 1,
            NCSTypeCode.OBJECT: 1,
            NCSTypeCode.EFFECT: 1,
            NCSTypeCode.EVENT: 1,
            NCSTypeCode.LOCATION: 1,
            NCSTypeCode.TALENT: 1,
            NCSTypeCode.VECTOR: 3,  # 12 bytes = 3 words
        }

        if self._type_code in size_map:
            return size_map[self._type_code]

        # Struct size must be determined externally
        if self._type_code == NCSTypeCode.STRUCT:
            msg = "Struct size must be determined by context"
            raise ValueError(msg)

        # Unknown type
        msg = f"Cannot determine size of type code: {self._type_code}"
        raise ValueError(msg)

    @property
    def type_code(self) -> NCSTypeCode:
        """Get the type code."""
        return self._type_code

    @property
    def size(self) -> int:
        """Get the size in 4-byte words."""
        return self._size

    @property
    def byte_size(self) -> int:
        """Get the size in bytes."""
        return self._size * 4

    def is_typed(self) -> bool:
        """Check if this is a valid typed value (not invalid/stack)."""
        return self._type_code not in {NCSTypeCode.INVALID, NCSTypeCode.STACK}

    def is_numeric(self) -> bool:
        """Check if this is a numeric type (int or float)."""
        return self._type_code in {NCSTypeCode.INTEGER, NCSTypeCode.FLOAT}

    def is_engine_type(self) -> bool:
        """Check if this is an engine-specific type."""
        return self._type_code in {
            NCSTypeCode.EFFECT,
            NCSTypeCode.EVENT,
            NCSTypeCode.LOCATION,
            NCSTypeCode.TALENT,
        }

    def to_string(self) -> str:
        """Convert type to its string representation.

        Returns:
        -------
            str: String representation (e.g., "int", "float")
        """
        type_names = {
            NCSTypeCode.NONE: "void",
            NCSTypeCode.STACK: "stack",
            NCSTypeCode.INTEGER: "int",
            NCSTypeCode.FLOAT: "float",
            NCSTypeCode.STRING: "string",
            NCSTypeCode.OBJECT: "object",
            NCSTypeCode.EFFECT: "effect",
            NCSTypeCode.EVENT: "event",
            NCSTypeCode.LOCATION: "location",
            NCSTypeCode.TALENT: "talent",
            NCSTypeCode.VECTOR: "vector",
            NCSTypeCode.STRUCT: "struct",
            NCSTypeCode.INVALID: "invalid",
            # Compound types
            NCSTypeCode.INTINT: "intint",
            NCSTypeCode.FLOATFLOAT: "floatfloat",
            NCSTypeCode.OBJECTOBJECT: "objectobject",
            NCSTypeCode.STRINGSTRING: "stringstring",
            NCSTypeCode.STRUCTSTRUCT: "structstruct",
            NCSTypeCode.INTFLOAT: "intfloat",
            NCSTypeCode.FLOATINT: "floatint",
            NCSTypeCode.EFFECTEFFECT: "effecteffect",
            NCSTypeCode.EVENTEVENT: "eventevent",
            NCSTypeCode.LOCLOC: "locloc",
            NCSTypeCode.TALTAL: "taltal",
            NCSTypeCode.VECTORVECTOR: "vectorvector",
            NCSTypeCode.VECTORFLOAT: "vectorfloat",
            NCSTypeCode.FLOATVECTOR: "floatvector",
        }

        return type_names.get(self._type_code, "unknown")

    def get_element(self, position: int) -> Self:
        """Get element at position (for struct/vector support).

        Args:
        ----
            position: Position index (1-based)

        Returns:
        -------
            NCSType: Type at position

        Raises:
        ------
            ValueError: If position is invalid for this type
        """
        if position != 1:
            msg = f"Position {position} > 1 for non-struct type"
            raise ValueError(msg)
        return self

    def get_component_type(self) -> NCSTypeCode:
        """Get the component type for compound types.

        Returns:
        -------
            NCSTypeCode: Component type code

        Examples:
        --------
            INTINT -> INTEGER
            FLOATFLOAT -> FLOAT
            VECTOR -> FLOAT (each component is float)
        """
        if self._type_code in {NCSTypeCode.INTINT, NCSTypeCode.INTFLOAT}:
            return NCSTypeCode.INTEGER
        if self._type_code in {NCSTypeCode.FLOATFLOAT, NCSTypeCode.FLOATINT, NCSTypeCode.FLOATVECTOR}:
            return NCSTypeCode.FLOAT
        if self._type_code in {NCSTypeCode.STRINGSTRING}:
            return NCSTypeCode.STRING
        if self._type_code in {NCSTypeCode.OBJECTOBJECT}:
            return NCSTypeCode.OBJECT
        if self._type_code == NCSTypeCode.VECTOR:
            return NCSTypeCode.FLOAT  # Vector components are floats
        return self._type_code

    def get_left_type(self) -> NCSTypeCode:
        """Get the left operand type for compound binary operations.

        Returns:
        -------
            NCSTypeCode: Left operand type code
        """
        if self._type_code in {NCSTypeCode.INTINT, NCSTypeCode.INTFLOAT}:
            return NCSTypeCode.INTEGER
        if self._type_code in {NCSTypeCode.FLOATFLOAT, NCSTypeCode.FLOATINT, NCSTypeCode.FLOATVECTOR}:
            return NCSTypeCode.FLOAT
        if self._type_code in {NCSTypeCode.STRINGSTRING}:
            return NCSTypeCode.STRING
        if self._type_code in {NCSTypeCode.OBJECTOBJECT}:
            return NCSTypeCode.OBJECT
        if self._type_code in {NCSTypeCode.EFFECTEFFECT}:
            return NCSTypeCode.EFFECT
        if self._type_code in {NCSTypeCode.EVENTEVENT}:
            return NCSTypeCode.EVENT
        if self._type_code in {NCSTypeCode.LOCLOC}:
            return NCSTypeCode.LOCATION
        if self._type_code in {NCSTypeCode.TALTAL}:
            return NCSTypeCode.TALENT
        if self._type_code in {NCSTypeCode.VECTORVECTOR, NCSTypeCode.VECTORFLOAT}:
            return NCSTypeCode.VECTOR
        if self._type_code in {NCSTypeCode.STRUCTSTRUCT}:
            return NCSTypeCode.STRUCT
        return self._type_code

    def get_right_type(self) -> NCSTypeCode:
        """Get the right operand type for compound binary operations.

        Returns:
        -------
            NCSTypeCode: Right operand type code
        """
        if self._type_code in {NCSTypeCode.INTINT, NCSTypeCode.FLOATINT}:
            return NCSTypeCode.INTEGER
        if self._type_code in {NCSTypeCode.FLOATFLOAT, NCSTypeCode.INTFLOAT, NCSTypeCode.VECTORFLOAT}:
            return NCSTypeCode.FLOAT
        if self._type_code in {NCSTypeCode.STRINGSTRING}:
            return NCSTypeCode.STRING
        if self._type_code in {NCSTypeCode.OBJECTOBJECT}:
            return NCSTypeCode.OBJECT
        if self._type_code in {NCSTypeCode.EFFECTEFFECT}:
            return NCSTypeCode.EFFECT
        if self._type_code in {NCSTypeCode.EVENTEVENT}:
            return NCSTypeCode.EVENT
        if self._type_code in {NCSTypeCode.LOCLOC}:
            return NCSTypeCode.LOCATION
        if self._type_code in {NCSTypeCode.TALTAL}:
            return NCSTypeCode.TALENT
        if self._type_code in {NCSTypeCode.VECTORVECTOR, NCSTypeCode.FLOATVECTOR}:
            return NCSTypeCode.VECTOR
        if self._type_code in {NCSTypeCode.STRUCTSTRUCT}:
            return NCSTypeCode.STRUCT
        return self._type_code

    def is_compound(self) -> bool:
        """Check if this is a compound type (e.g., INTINT, FLOATFLOAT).

        Returns:
        -------
            bool: True if compound type
        """
        return self._type_code in {
            NCSTypeCode.INTINT,
            NCSTypeCode.FLOATFLOAT,
            NCSTypeCode.OBJECTOBJECT,
            NCSTypeCode.STRINGSTRING,
            NCSTypeCode.STRUCTSTRUCT,
            NCSTypeCode.INTFLOAT,
            NCSTypeCode.FLOATINT,
            NCSTypeCode.EFFECTEFFECT,
            NCSTypeCode.EVENTEVENT,
            NCSTypeCode.LOCLOC,
            NCSTypeCode.TALTAL,
            NCSTypeCode.VECTORVECTOR,
            NCSTypeCode.VECTORFLOAT,
            NCSTypeCode.FLOATVECTOR,
        }

    def is_vector(self) -> bool:
        """Check if this is a vector type.

        Returns:
        -------
            bool: True if vector type
        """
        return self._type_code == NCSTypeCode.VECTOR

    def is_struct(self) -> bool:
        """Check if this is a struct type.

        Returns:
        -------
            bool: True if struct type
        """
        return self._type_code == NCSTypeCode.STRUCT

    def can_convert_to(self, target_type: NCSTypeCode) -> bool:
        """Check if this type can be converted to target type.

        Args:
        ----
            target_type: Target type code

        Returns:
        -------
            bool: True if conversion is possible
        """
        # Numeric types can convert between each other
        if self.is_numeric() and target_type in {NCSTypeCode.INTEGER, NCSTypeCode.FLOAT}:
            return True
        # Same type always works
        if self._type_code == target_type:
            return True
        # Vector to float (component access)
        return bool(self._type_code == NCSTypeCode.VECTOR and target_type == NCSTypeCode.FLOAT)

    @staticmethod
    def create_compound(left_type: NCSTypeCode, right_type: NCSTypeCode) -> NCSTypeCode:
        """Create a compound type code from two operand types.

        Args:
        ----
            left_type: Left operand type
            right_type: Right operand type

        Returns:
        -------
            NCSTypeCode: Compound type code

        Examples:
        --------
            INTEGER, INTEGER -> INTINT
            FLOAT, FLOAT -> FLOATFLOAT
            INTEGER, FLOAT -> INTFLOAT
        """
        if left_type == NCSTypeCode.INTEGER and right_type == NCSTypeCode.INTEGER:
            return NCSTypeCode.INTINT
        if left_type == NCSTypeCode.FLOAT and right_type == NCSTypeCode.FLOAT:
            return NCSTypeCode.FLOATFLOAT
        if left_type == NCSTypeCode.INTEGER and right_type == NCSTypeCode.FLOAT:
            return NCSTypeCode.INTFLOAT
        if left_type == NCSTypeCode.FLOAT and right_type == NCSTypeCode.INTEGER:
            return NCSTypeCode.FLOATINT
        if left_type == NCSTypeCode.STRING and right_type == NCSTypeCode.STRING:
            return NCSTypeCode.STRINGSTRING
        if left_type == NCSTypeCode.OBJECT and right_type == NCSTypeCode.OBJECT:
            return NCSTypeCode.OBJECTOBJECT
        if left_type == NCSTypeCode.EFFECT and right_type == NCSTypeCode.EFFECT:
            return NCSTypeCode.EFFECTEFFECT
        if left_type == NCSTypeCode.EVENT and right_type == NCSTypeCode.EVENT:
            return NCSTypeCode.EVENTEVENT
        if left_type == NCSTypeCode.LOCATION and right_type == NCSTypeCode.LOCATION:
            return NCSTypeCode.LOCLOC
        if left_type == NCSTypeCode.TALENT and right_type == NCSTypeCode.TALENT:
            return NCSTypeCode.TALTAL
        if left_type == NCSTypeCode.VECTOR and right_type == NCSTypeCode.VECTOR:
            return NCSTypeCode.VECTORVECTOR
        if left_type == NCSTypeCode.VECTOR and right_type == NCSTypeCode.FLOAT:
            return NCSTypeCode.VECTORFLOAT
        if left_type == NCSTypeCode.FLOAT and right_type == NCSTypeCode.VECTOR:
            return NCSTypeCode.FLOATVECTOR
        if left_type == NCSTypeCode.STRUCT and right_type == NCSTypeCode.STRUCT:
            return NCSTypeCode.STRUCTSTRUCT
        return NCSTypeCode.INVALID

    def __str__(self) -> str:
        """String representation of the type."""
        return self.to_string()

    def __repr__(self) -> str:
        """Developer representation of the type."""
        return f"NCSType({self._type_code.name})"

    def __eq__(self, other: object) -> bool:
        """Check equality with another type."""
        if isinstance(other, NCSType):
            return self._type_code == other._type_code
        if isinstance(other, (int, NCSTypeCode)):
            return self._type_code == other
        return NotImplemented  # type: ignore[no-any-return]

    def __hash__(self) -> int:
        """Hash for use in dictionaries/sets."""
        return hash(self._type_code)


def create_type(type_input: int | NCSTypeCode | str | NCSType) -> NCSType:
    """Create an NCSType from various input types.

    Args:
    ----
        type_input: Type specification

    Returns:
    -------
        NCSType: Created type object
    """
    if isinstance(type_input, NCSType):
        return type_input
    return NCSType(type_input)



from __future__ import annotations

from ctypes import POINTER, addressof, c_uint, c_void_p, cast as cast_with_ctypes, pointer
from enum import IntEnum
from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from ctypes import _CData, _Pointer as PointerType

if not TYPE_CHECKING:
    PointerType = POINTER(c_uint).__class__


PREVENT_GC: list[Any] = []


class PointerHandlerMode(IntEnum):
    CTYPES_CAST = 0
    CTYPES_ARRAY = 1
    NUMPY = 2


def create_pointer(
    address: CPointer,
    mode: PointerHandlerMode = PointerHandlerMode.CTYPES_CAST,
) -> CPointer:
    if mode is PointerHandlerMode.CTYPES_CAST:
        new_pointer = CPointer(addressof(address))
    elif mode is PointerHandlerMode.CTYPES_ARRAY:
        pointer_array = (CPointer * 1)(address)
        new_pointer = CPointer(addressof(pointer_array))
        PREVENT_GC.append(pointer_array)
    elif mode is PointerHandlerMode.NUMPY:
        import numpy as np

        np_array = np.array([address.value], dtype=np.uintp)
        new_pointer = CPointer(np_array.ctypes.data)
        PREVENT_GC.append(np_array)
    PREVENT_GC.append(new_pointer)
    return new_pointer


def follow_pointer(
    address: CPointer | _CData,
    mode: PointerHandlerMode = PointerHandlerMode.CTYPES_CAST,
) -> CPointer:
    if not isinstance(address, c_void_p):
        address = c_void_p.from_address(addressof(address))
    if address.value == 0xDDDDDDDDDDDDDDDD:  # noqa: PLR2004
        raise OSError("INVALID pointer (Unallocated memory)")
    if not address.value:
        raise OSError(f"NULL pointer (address.value == {address.value})")
    if mode is PointerHandlerMode.CTYPES_CAST:
        return cast_with_ctypes(address, POINTER(CPointer)).contents
    if mode is PointerHandlerMode.CTYPES_ARRAY:
        # I have no early idea why this throws an exception... it's the same code?
        # deref_obj = CPointer.from_address(address.value)
        # dereferenced_address = deref_obj.value
        dereferenced_address = CPointer.from_address(address.value).value
        print(f"follow_address: Following address {hex(address.value)} to {hex(dereferenced_address or 0)}")
        assert dereferenced_address is not None, f"Dereferenced address should not be None: {hex(address.value)}"
        return CPointer(dereferenced_address)
    if mode is PointerHandlerMode.NUMPY:
        import numpy as np

        np_array = np.ctypeslib.as_array((c_uint * 1).from_address(address.value or 0))
        return CPointer(np_array[0])
    raise ValueError("Invalid/unsupported mode")


class CPointer(c_void_p):
    """A custom pointer type mirroring ctypes pointer functionality supporting original object return."""

    resolve: Callable[..., _CData]

    def __init__(self, address: int, *args, reinit: bool = False):
        if not reinit:
            super().__init__(address)
        self._address: int = address
        PREVENT_GC.append(self)

    def __hash__(self) -> int:
        return hash((self._address, self.value))

    def follow(self, *, mode: PointerHandlerMode = PointerHandlerMode.CTYPES_CAST) -> CPointer:
        return follow_pointer(self, mode=mode)

    def new_pointer(self, *, mode: PointerHandlerMode = PointerHandlerMode.CTYPES_CAST) -> CPointer:
        return create_pointer(self, mode=mode)


def adjust_pointer_depth(  # noqa: C901, ANN201
    obj: _CData | PointerType,
    pointer_depth: int,
    *,
    mode: PointerHandlerMode = PointerHandlerMode.CTYPES_CAST,
    max_depth: int = 255,
):
    """Create a custom pointer to the given object with the specified depth. Negative values mean resolving/following pointers.

    Args:
        obj: The object to create the pointer to.
        pointer_depth (int): The depth of pointers to create or follow.

    Returns:
        CPointer: A custom pointer to the object with the specified depth.
            or obj: the real object at the end of the pointer trail.
    """
    # Step 1: Dereference the pointers to get the actual data.
    actual_obj = obj
    while isinstance(actual_obj, PointerType):
        actual_obj = actual_obj.contents
        pointer_depth += 1
    # We now have the actual object.
    assert not isinstance(actual_obj, PointerType)
    if not pointer_depth:
        return actual_obj  # pointer_depth of 0 doesn't make sense past this point.

    actual_obj_type = actual_obj.__class__
    orig_addr = addressof(actual_obj)

    def resolve(self: CPointer) -> _CData:
        """Follow the pointer to return the original object."""
        current = self
        cur_depth = 1
        last_ptr = None
        while addressof(current) != orig_addr:
            print(
                "Resolving to type:",
                actual_obj_type.__name__,
                "Address of Current:",
                hex(addressof(current)),
                "Current Address:",
                hex(0 if current.value is None else current.value),
                "Current Depth Level:",
                cur_depth,
            )
            last_ptr = current
            current = follow_pointer(current, mode=mode)
            assert isinstance(current, (PointerType, c_void_p)), current.__class__.__name__
            cur_depth += 1
            if cur_depth >= max_depth:
                raise RuntimeError(f"Max pointer depth {max_depth} reached!")
        assert last_ptr is not None
        return cast_with_ctypes(last_ptr, POINTER(actual_obj_type)).contents

    CPointer.resolve = resolve

    # Initialize pointer storage and pointer
    current_ptr: CPointer = CPointer(orig_addr)
    depth = abs(pointer_depth)

    # Adjust pointer based on depth
    for _ in range(depth):
        current_ptr = create_pointer(current_ptr, mode=mode) if pointer_depth > 0 else follow_pointer(current_ptr, mode=mode)
        PREVENT_GC.append(current_ptr)

    retptr = CPointer(addressof(current_ptr))
    PREVENT_GC.append(retptr)

    deref_obj = current_ptr
    for _ in range(depth + 1):
        deref_obj = follow_pointer(deref_obj, mode=mode)
    deref_addr = addressof(deref_obj)

    print(f"levels: {pointer_depth}, deref: {hex(deref_addr)}, orig: {hex(orig_addr)}")
    assert orig_addr == deref_addr, f"Original address DOES NOT match dereferenced address! ({hex(orig_addr)} == {hex(deref_addr)})"
    return retptr


if __name__ == "__main__":
    print(adjust_pointer_depth(c_uint(5), 2).resolve())
    print(adjust_pointer_depth(c_uint(5), 1).resolve())
    print(adjust_pointer_depth(c_uint(5), 0))
    print(adjust_pointer_depth(c_void_p(addressof(c_uint(5))), -1))
    print(adjust_pointer_depth(pointer(c_uint(5)), -1))
    print(adjust_pointer_depth(c_uint(5), 5).resolve())
    print(adjust_pointer_depth(pointer(c_uint(5)), 0))
    # print(adjust_pointer_depth(pointer(c_uint(5)), 1).resolve())  # Seems to only fail in 'run without debugging' mode...
